/*
	APL Restful ARDBC Plugin
    Copyright 2015-2018  LJ LongWing - A Programming Legacy (http://remedylegacy.com)

	Licensed under the Apache License, Version 2.0 (the "License");
	you may not use this file except in compliance with the License.
	You may obtain a copy of the License at

		http://www.apache.org/licenses/LICENSE-2.0

	Unless required by applicable law or agreed to in writing, software
	distributed under the License is distributed on an "AS IS" BASIS,
	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
	See the License for the specific language governing permissions and
	limitations under the License.
 */

package com.apl;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.net.Authenticator;
import java.net.CookieHandler;
import java.net.CookieManager;
import java.net.CookiePolicy;
import java.net.HttpURLConnection;
import java.net.InetSocketAddress;
import java.net.MalformedURLException;
import java.net.PasswordAuthentication;
import java.net.ProtocolException;
import java.net.Proxy;
import java.net.SocketTimeoutException;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLConnection;
import java.nio.charset.StandardCharsets;
import java.security.KeyManagementException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.KeyManagerFactory;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSession;
import javax.net.ssl.TrustManager;
import javax.net.ssl.X509TrustManager;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathFactory;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import com.bmc.arsys.api.ARErrors;
import com.bmc.arsys.api.ARException;
import com.bmc.arsys.api.ARServerUser;
import com.bmc.arsys.api.ArithmeticOrRelationalOperand;
import com.bmc.arsys.api.AttachmentFieldLimit;
import com.bmc.arsys.api.AttachmentValue;
import com.bmc.arsys.api.CharacterFieldLimit;
import com.bmc.arsys.api.Constants;
import com.bmc.arsys.api.DataType;
import com.bmc.arsys.api.DateTimeFieldLimit;
import com.bmc.arsys.api.Entry;
import com.bmc.arsys.api.EntryListFieldInfo;
import com.bmc.arsys.api.EntryListInfo;
import com.bmc.arsys.api.FieldLimit;
import com.bmc.arsys.api.IntegerFieldLimit;
import com.bmc.arsys.api.OutputInteger;
import com.bmc.arsys.api.QualifierInfo;
import com.bmc.arsys.api.RelationalOperationInfo;
import com.bmc.arsys.api.SortInfo;
import com.bmc.arsys.api.StatisticsResultInfo;
import com.bmc.arsys.api.Timestamp;
import com.bmc.arsys.api.Value;
import com.bmc.arsys.api.VendorForm;
import com.bmc.arsys.pluginsvr.plugins.ARDBCPlugin;
import com.bmc.arsys.pluginsvr.plugins.ARPluginContext;
import com.bmc.arsys.pluginsvr.plugins.ARVendorField;
import com.bmc.arsys.util.ARUtilEgcp;

public class RestfulARDBCPlugin extends ARDBCPlugin {
	String version = "4.12";

	private volatile static String tableForm, fieldForm, cacheForm, configForm, functionForm, remedyServerName,
			proxyServer, proxyUser, proxyPass, keyStoreAlgorithm, keyStoreFormat, keyStoreFile, keyStorePass,
			keyStoreProtocol, templateFieldDelimiter, arrayDelimiter;
	private volatile static int remedyServerPort, keepCacheSeconds, keepResultSeconds, timeout, proxyPort;
	private volatile static boolean alwaysLogDetails;

	private Map<String, String> globalElementMap = new TreeMap<String, String>();
	private Map<String, String> recordElementMap = new TreeMap<String, String>();
	private Map<String, APLField> aplFieldMap = new HashMap<String, APLField>();

	private List<Map<String, String>> recordElementList = new ArrayList<>();
	private List<String> fieldOptionalList = new ArrayList<>();
	private List<String> fieldLocalFilterList = new ArrayList<>();
	private List<Integer> skipLevelList = new ArrayList<>();
	public static List<byte[]> attByteList  = new ArrayList<>();

	private static boolean remedyServer = false;
	private boolean debug = false;
	private boolean bypassCertCheck = false;
	private boolean bypassProxy = false;
	private boolean allowAllReturnCodes = false;

	public ARServerUser conn = null;

	private String urlS = "";
	private String allowAllReturnCodesS;
	private String tableRequestID;
	private String tableGUID;
	private String tableURL;
	private String tableURLSuffix;
	private String tableRecordIndicatorElement;
	private String tableBasicUser;
	private String tableBasicPass;
	private String tableRemedyServer;
	private String tableCustomHeader;
	private String tableTimeoutS;
	private String tableGetMethod;
	private String tableGetTemplate;
	private String tableCreateMethod;
	private String tableCreateTemplate;
	private String tableSetMethod;
	private String tableSetTemplate;
	private String tableDeleteMethod;
	private String tableDeleteTemplate;
	private String tableCreateRequestIDElement;
	private String tableBypassSSLChecks;
	private String tableBypassProxy;
	private String tableSkipElementLevels;
	private String encodedUrl;
	private String urlRoot;
	private String remedyToken = null;

	private int tableGetResponseCode;
	private int tableCreateResponseCode;
	private int tableSetResponseCode;
	private int tableDeleteResponseCode;
	private int tableRecordIndicatorLevel = -1;
	private int responseCode = 0;

	private URL url = null;

	private URLConnection webConn = null;

	private Proxy proxy;

	private Authenticator basicAuth = null;

	public Entry pluginFunctions(Map<String, Object> fieldNameValueMap, ARPluginContext context, long transid) {
		checkConnection(context);

		String function = "";
		Entry functionEntry = new Entry();

		functionEntry.put(APLConstants.fidFunctionFunction, new Value((String)fieldNameValueMap.get("Function")));
		functionEntry.put(APLConstants.fidFunctionInput1, new Value((String)fieldNameValueMap.get("Input1")));
		functionEntry.put(APLConstants.fidFunctionInput2, new Value((String)fieldNameValueMap.get("Input2")));
		functionEntry.put(APLConstants.fidFunctionInput3, new Value((String)fieldNameValueMap.get("Input3")));
		functionEntry.put(APLConstants.fidFunctionInput4, new Value((String)fieldNameValueMap.get("Input4")));
		functionEntry.put(APLConstants.fidFunctionInput5, new Value((String)fieldNameValueMap.get("Input5")));
		functionEntry.put(APLConstants.fidFunctionInput6, new Value((String)fieldNameValueMap.get("Input6")));
		functionEntry.put(APLConstants.fidFunctionInput7, new Value((String)fieldNameValueMap.get("Input7")));

		if (fieldNameValueMap.containsKey("Function"))
			function = (String)fieldNameValueMap.get("Function");

		String request = "";

		debug = true;

		if (!function.isEmpty()) {
			StringBuilder sb = new StringBuilder();

			if (function.equals("ShowCache")) {
				request = Functions.generateEntryID(1, transid);

				List<Entry> entryList = getListCacheEntry(context, null, null, null, "Cache", false);

				String tableName = "";
				String url = "";
				long timestamp = 0;
				for (Entry entry : entryList) {
					String curTableName = entry.get(APLConstants.fidCacheTableName).toString();
					if (!curTableName.equals(tableName)) {
						tableName = curTableName;
						sb.append("Table: " + tableName + "\r\n");
					}

					String curURL = entry.get(APLConstants.fidCacheURL).toString();
					if (!curURL.equals(url)) {
						url = curURL;
						sb.append("-URL: " + url + "\r\n");
					}

					long curTimestamp = Long
							.valueOf(entry.get(APLConstants.fidCacheEntryID).toString().substring(0, 10));
					if (curTimestamp != timestamp) {
						timestamp = curTimestamp;
						sb.append("-Retrieve at: " + new Date(timestamp) + " (" + timestamp + ")\r\n");
					}

					Entry recordEntry = Functions.deserializeEntryObject(context,
							entry.get(APLConstants.fidCacheEntryObject).toString());
					sb.append("--Entry ID: " + recordEntry.getEntryId() + "\r\n");
					for (Integer fieldID : recordEntry.keySet()) {
						sb.append("---FieldID/Value: " + fieldID + ": " + recordEntry.get(fieldID) + "\r\n");
					}
				}

				if (sb.toString().isEmpty())
					sb.append("Cache is currently empty");
			} else if (function.equals("ShowResults")) {
				request = Functions.generateEntryID(2, transid);

				List<Entry> entryList = getListCacheEntry(context, null, null, null, "ResultList", false);

				String tableName = "";
				String user = "";
				long timestamp = 0;
				for (Entry entry : entryList) {
					String curTableName = entry.get(APLConstants.fidCacheTableName).toString();
					if (!curTableName.equals(tableName)) {
						tableName = curTableName;
						sb.append("Table: " + tableName + "\r\n");
					}

					String curUser = entry.get(APLConstants.fidCacheUser).toString();
					if (!curUser.equals(user)) {
						user = curUser;
						sb.append("-Retrieved by " + user + "\r\n");
					}

					long curTimestamp = Long
							.valueOf(entry.get(APLConstants.fidCacheEntryID).toString().substring(0, 10));
					if (curTimestamp != timestamp) {
						timestamp = curTimestamp;
						sb.append("-Retrieve on: " + new Date(timestamp) + " (" + timestamp + ")\r\n");
					}

					Entry recordEntry = Functions.deserializeEntryObject(context,
							entry.get(APLConstants.fidCacheEntryObject).toString());
					sb.append("--Entry ID: " + recordEntry.getEntryId() + "\r\n");
					for (Integer fieldID : recordEntry.keySet()) {
						sb.append("---FieldID/Value: " + fieldID + ": " + recordEntry.get(fieldID) + "\r\n");
					}
				}

				if (sb.toString().isEmpty())
					sb.append("Results List cache is currently empty");
			} else if (function.equals("GenFields")) {
				request = Functions.generateEntryID(3, transid);

				String tableGUID = (String)fieldNameValueMap.get("Input1");
				String tableName = (String)fieldNameValueMap.get("Input2");
				urlS = (String)fieldNameValueMap.get("Input3");
				String sampleMethod = (String)fieldNameValueMap.get("Input4");
				String sampleBody = (String)fieldNameValueMap.get("Input5");

				try {
					getTableMetaData(context, tableName, null, fieldNameValueMap, null, true, transid);
				} catch (ARException e) {
					sb.append("ARException getting table metadata: " + e.getMessage());
					return returnEntryObject(functionEntry, request, function, sb);
				}

				tableRecordIndicatorElement = "";
				tableRecordIndicatorLevel = -1;

				// ok, it's time to encode the URL for proper transmission
				try {
					encodedUrl = Functions.encodeURL(urlS);
				} catch (MalformedURLException e) {
					sb.append("URL: " + urlS + "\r\nMalformedURLException: " + e.getMessage());
					returnEntryObject(functionEntry, request, function, sb);
				} catch (URISyntaxException e) {
					sb.append("URL: " + urlS + "\r\nURISyntaxException: " + e.getMessage());
					returnEntryObject(functionEntry, request, function, sb);
				}

				try {
					url = new URL(encodedUrl);
				} catch (MalformedURLException e) {
					sb.append("URL: " + encodedUrl + "\r\nMalformedURLException: " + e.getMessage());
					returnEntryObject(functionEntry, request, function, sb);
				}

				try {
					makeRestCall(context, null, null, sampleMethod, sampleBody, 0, null, transid);
				} catch (ARException e) {
					sb.append("ARException making call: " + e.getMessage());
					return returnEntryObject(functionEntry, request, function, sb);
				}

				InputStream is = null;

				try {
					is = webConn.getInputStream();
				} catch (IOException e) {
					is = ((HttpURLConnection) webConn).getErrorStream();
					//sb.append("IOException getting web response: " + e.getMessage());
					//return returnEntryObject(functionEntry, request, function, sb);
				}

				String output = Functions.processInputStream(is, Functions.getCharacterSet(webConn.getContentType()))
						.trim();
				if (output.startsWith("{") || output.startsWith("[")) {
					JSONParser handler = null;
					handler = new JSONParser(null, tableRecordIndicatorElement, tableRecordIndicatorLevel,
							skipLevelList);
					handler.parse(output, arrayDelimiter);

					globalElementMap = handler.getGlobalElementMap();
				} else if (output.startsWith("<")) {
					is = Functions.stringToIS(output, webConn.getContentType());
					SAXParserFactory factory = SAXParserFactory.newInstance();
					XMLParser handler = new XMLParser(tableRecordIndicatorElement, null, null);

					try {
						SAXParser parser = factory.newSAXParser();

						parser.parse(is, handler);
					} catch (SAXException e) {
						sb.append("SAXException: " + e.getMessage());
						return returnEntryObject(functionEntry, request, function, sb);
					} catch (ParserConfigurationException e) {
						sb.append("ParserConfigurationException: " + e.getMessage());
						return returnEntryObject(functionEntry, request, function, sb);
					} catch (IOException e) {
						sb.append("IOException parsing XML: " + e.getMessage());
						return returnEntryObject(functionEntry, request, function, sb);
					}

					globalElementMap = handler.getGlobalElementMap(arrayDelimiter);
				}
				if (globalElementMap == null)
					globalElementMap = new TreeMap<String, String>();
				
				globalElementMap.putAll(Functions.generateHeaderElements(webConn.getHeaderFields(), context, debug));
				outputElements(context);

				((HttpURLConnection) webConn).disconnect();

				if (globalElementMap != null) {
					for (String element : globalElementMap.keySet()) {
						QualifierInfo fieldQual = Functions.getQualifierInfo(APLConstants.fidFieldName, element,
								RelationalOperationInfo.AR_REL_OP_EQUAL);

						QualifierInfo qual = new QualifierInfo(QualifierInfo.AR_COND_OP_AND,
								Functions.getActiveGUIDQual(tableGUID), fieldQual);

						List<EntryListInfo> entryList;
						try {
							entryList = conn.getListEntry(fieldForm, qual, 0, 0, null, null, false, null);

							if (entryList.size() == 0) {
								Entry entry = new Entry();

								entry.put(APLConstants.fidGuid, new Value(tableGUID));
								entry.put(APLConstants.fidFieldName, new Value(element));

								conn.createEntry(fieldForm, entry);

								sb.append("Created " + fieldForm + " record for element '" + element + "'\r\n");
							}
						} catch (ARException e) {
							sb.append("ARException processing Elements: " + e.getMessage());
							return returnEntryObject(functionEntry, request, function, sb);
						}
					}
				}
				if (sb.toString().isEmpty())
					sb.append("No Fields Found to Add");
			} else if (function.equals("CurrentSettings")) {
				request = Functions.generateEntryID(6, transid);

				sb.append("APLRestfulARDBCPlugin version " + version + System.lineSeparator());
				sb.append("AlwaysLogDetails: " + alwaysLogDetails + System.lineSeparator());
				sb.append("ArrayDelimiter: " + arrayDelimiter + System.lineSeparator());
				sb.append("CacheForm: " + cacheForm + System.lineSeparator());
				sb.append("ConfigForm: " + configForm + System.lineSeparator());
				sb.append("FieldForm: " + fieldForm + System.lineSeparator());
				sb.append("FunctionForm: " + functionForm + System.lineSeparator());
				sb.append("KeepCacheSeconds: " + keepCacheSeconds + System.lineSeparator());
				sb.append("KeepResultSeconds: " + keepResultSeconds + System.lineSeparator());
				sb.append("KeyStoreAlgorithm: " + keyStoreAlgorithm + System.lineSeparator());
				sb.append("KeyStoreFile: " + keyStoreFile + System.lineSeparator());
				sb.append("KeyStoreFormat: " + keyStoreFormat + System.lineSeparator());
				sb.append("KeyStoreProtocol: " + keyStoreProtocol + System.lineSeparator());
				sb.append("ProxyPass: " + proxyPass + System.lineSeparator());
				sb.append("ProxyPort: " + proxyPort + System.lineSeparator());
				sb.append("ProxyServer: " + proxyServer + System.lineSeparator());
				sb.append("ProxyUser: " + proxyUser + System.lineSeparator());
				sb.append("RemedyServerName: " + remedyServerName + System.lineSeparator());
				sb.append("RemedyServerPort: " + remedyServerPort + System.lineSeparator());
				sb.append("TableForm: " + tableForm + System.lineSeparator());
				sb.append("TemplateFieldDelimiter: " + templateFieldDelimiter + System.lineSeparator());
				sb.append("Timeout: " + timeout + System.lineSeparator());

			} else {
				request = Functions.generateEntryID(0, transid);
			}

			return returnEntryObject(functionEntry, request, function, sb);
		}

		return null;
	}

	private Entry returnEntryObject(Entry entry, String entryID, String function, StringBuilder sb) {
		entry.setEntryId(entryID);
		entry.put(1, new Value(entryID));
		entry.put(APLConstants.fidFunctionFunction, new Value(function));
		entry.put(APLConstants.fidFunctionOutput, new Value(sb.toString()));
		return entry;
	}

	public void cancel(ARPluginContext context, long transid) throws ARException {
		Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "cancel(): " + transid);
		super.cancel(context, transid);
	}

	public void commit(ARPluginContext context, long transid) throws ARException {
		Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "commit(): " + transid);
		super.commit(context, transid);
	}

	public String createEntry(ARPluginContext context, String tableName, List<ARVendorField> fields, long transid,
			Entry newEntry) throws ARException {
		Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "createEntry on " + tableName);
		getConfigDetails(context, false);

		globalElementMap = new TreeMap<String, String>();

		//System.out.println("createEntry: call processEntry");
		Map<String, Object> fieldNameValueMap = Functions.processEntry(newEntry, fields, aplFieldMap);

		fieldOptionalList = new ArrayList<String>();
		//System.out.println("createEntry: call getTableMetaData");
		List<Entry> tableMetaDataList = getTableMetaData(context, tableName, fields, fieldNameValueMap, "CREATE", false,
				transid);

		fieldNameValueMap = Functions.processEntry(newEntry, fields, aplFieldMap);
		
		if (tableMetaDataList != null) {
			String message = Functions.getErrorString(fields, tableMetaDataList);
			throw new ARException(Constants.AR_RETURN_ERROR, ARErrors.AR_PLUGIN_ERROR_UNKOWN, message);
		}

		String createTemplate = Functions.replaceTemplateValues(tableCreateTemplate, fieldNameValueMap,
				fieldOptionalList, aplFieldMap, templateFieldDelimiter);

		List<Entry> restResultsList = makeRestCall(context, tableName, fields, tableCreateMethod, createTemplate,
				tableCreateResponseCode, fieldNameValueMap, transid);

		if (restResultsList != null) {
			String message = Functions.getErrorString(fields, restResultsList);
			throw new ARException(Constants.AR_RETURN_ERROR, ARErrors.AR_PLUGIN_ERROR_UNKOWN, message);
		}

		String requestID = "SpecifyElement";

		if (tableCreateRequestIDElement != null && !tableCreateRequestIDElement.isEmpty())
			requestID = "";

		InputStream is = null;

		try {
			is = webConn.getInputStream();
		} catch (IOException e) {
			is = ((HttpURLConnection) webConn).getErrorStream();
		}

		String output = Functions.processInputStream(is, null);
		if (output.startsWith("<")) { // XML
			is = Functions.stringToIS(output, webConn.getContentType());
			SAXParserFactory factory = SAXParserFactory.newInstance();
			XMLParser handler = new XMLParser(tableRecordIndicatorElement, skipLevelList, fields);

			try {
				SAXParser parser = factory.newSAXParser();

				parser.parse(is, handler);
			} catch (SAXException e) {
				String message = "SAXException: " + e.getMessage();
				throw new ARException(Constants.AR_RETURN_ERROR, ARErrors.AR_PLUGIN_ERROR_UNKOWN, message);
			} catch (ParserConfigurationException e) {
				String message = "ParserConfigurationException: " + e.getMessage();
				throw new ARException(Constants.AR_RETURN_ERROR, ARErrors.AR_PLUGIN_ERROR_UNKOWN, message);
			} catch (IOException e) {
				String message = "IOException parsing XML: " + e.getMessage();
				throw new ARException(Constants.AR_RETURN_ERROR, ARErrors.AR_PLUGIN_ERROR_UNKOWN, message);
			}

			globalElementMap = handler.getGlobalElementMap(arrayDelimiter);
			recordElementMap = handler.getRecordElementMap(arrayDelimiter);
			recordElementList = handler.getRecordElementList();
		} else if (output.startsWith("{") || output.startsWith("[")) {
			JSONParser handler = null;
			handler = new JSONParser(fields, tableRecordIndicatorElement, tableRecordIndicatorLevel, skipLevelList);
			handler.parse(output, arrayDelimiter);

			globalElementMap = handler.getGlobalElementMap();
			recordElementMap = handler.getRecordElementMap();
			recordElementList = handler.getRecordElementList();
		} // end if for XML/JSON

		if (globalElementMap == null)
			globalElementMap = new TreeMap<String, String>();

		globalElementMap.putAll(Functions.generateHeaderElements(webConn.getHeaderFields(), context, debug));

		((HttpURLConnection) webConn).disconnect();

		if (requestID.isEmpty() && globalElementMap != null) {
			if (globalElementMap.containsKey(tableCreateRequestIDElement)) {
				requestID = globalElementMap.get(tableCreateRequestIDElement);
			} else {
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "ReqID is Empty & GlobalElementMap is not, below are the contents of the map:");
				for (String elementName : globalElementMap.keySet()) {
					Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, elementName + ": " + globalElementMap.get(elementName));
				}
			}
		} else {
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "Either RequestID is empty, or globalElementMap is:");
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "RequestID: " + requestID);
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "GlobalElementMap: " + globalElementMap);
		}

		if (requestID.isEmpty() && recordElementMap != null) {
			if (recordElementMap.containsKey(tableCreateRequestIDElement)) {
				requestID = recordElementMap.get(tableCreateRequestIDElement);
			}
		}

		if (requestID.isEmpty() && recordElementMap != null && recordElementMap.containsKey(tableCreateRequestIDElement)
				&& recordElementList.size() > 0) {
			requestID = recordElementList.get(0).get(tableCreateRequestIDElement);
		}

		if (!requestID.equals("SpecifyElement") && !requestID.isEmpty() && debug) {
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "Record Created with Request ID '" + requestID + "'");
		}

		String s = (String)fieldNameValueMap.get("OutputUnusedElements");
		if (s != null && !s.isEmpty()) {
			if (globalElementMap != null) {
				for (String element : globalElementMap.keySet()) {
					Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "Returned Element/Value: " + element + ":" + globalElementMap.get(element));
				}
			}
			if (recordElementMap != null) {
				for (String element : recordElementMap.keySet()) {
					Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "Returned Element/Value: " + element + ":" + recordElementMap.get(element));
				}
			}
		}

		return requestID;
	}

	public void deleteEntry(ARPluginContext context, String tableName, List<ARVendorField> fields, long transid,
			String entryId) throws ARException {
		Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "deleteEntry on " + tableName + " for entryId " + entryId);
		getConfigDetails(context, false);

		globalElementMap = new TreeMap<String, String>();

		int[] fieldIDArray = new int[fields.size()];

		for (int i = 0; i < fields.size(); i++) {
			int fieldID = fields.get(i).getFieldId();
			fieldIDArray[i] = fieldID;
		}

		Entry entry = getEntry(context, tableName, fields, transid, entryId, fieldIDArray);

		Map<String, Object> fieldNameValueMap = Functions.processEntry(entry, fields, aplFieldMap);

		fieldOptionalList = new ArrayList<String>();
		List<Entry> tableMetaDataList = getTableMetaData(context, tableName, fields, fieldNameValueMap, "DELETE", false,
				transid);

		if (tableMetaDataList != null) {
			String message = Functions.getErrorString(fields, tableMetaDataList);
			throw new ARException(Constants.AR_RETURN_ERROR, ARErrors.AR_PLUGIN_ERROR_UNKOWN, message);
		}

		String deleteTemplate = Functions.replaceTemplateValues(tableDeleteTemplate, fieldNameValueMap,
				fieldOptionalList, aplFieldMap, templateFieldDelimiter);

		List<Entry> restResultsList = makeRestCall(context, tableName, fields, tableDeleteMethod, deleteTemplate,
				tableDeleteResponseCode, fieldNameValueMap, transid);

		if (restResultsList != null) {
			String message = Functions.getErrorString(fields, restResultsList);
			throw new ARException(Constants.AR_RETURN_ERROR, ARErrors.AR_PLUGIN_ERROR_UNKOWN, message);
		}

		((HttpURLConnection) webConn).disconnect();
	}

	public AttachmentValue getBLOB(ARPluginContext context, String tableName, List<ARVendorField> fields, long transid,
			String entryId, int fieldId) throws ARException {
		Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "getBlob() on " + tableName + " on entryId " + entryId);
		getConfigDetails(context, false);

		int[] fieldIds = { fieldId };
		Entry entry = getEntry(context, tableName, fields, transid, entryId, fieldIds);

		Value value = entry.get(fieldId);

		if (value.getDataType() == DataType.ATTACHMENT
				|| value.getDataType().toInt() == Constants.AR_DATA_TYPE_ATTACH) {
			AttachmentValue av = (AttachmentValue) value.getValue();
			String attFileName = av.getName();
			String attURL = av.getValueFileName();
			return getAttachment(context, attFileName, attURL, fields, tableName, transid);
		} else {
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "getBlob() is not an Attachment field: " + value.getDataType());
		}

		return null;
	}

	public Entry getEntry(ARPluginContext context, String tableName, List<ARVendorField> fields, long transid,
			String entryId, int[] fieldIds) throws ARException {
		Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "getEntry() on " + tableName + " for entryId " + entryId);
		getConfigDetails(context, false);

		String userName = Functions.setUserName(context);

		// if (entryId.length() == 15) {
		Entry cacheEntry = manageCacheEntry(context, null, null, tableName, entryId, "ResultList", userName, null,
				"Get", false);
		if (cacheEntry != null) {
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "found ResultList entry");
			return cacheEntry;
		} else {
			cacheEntry = manageCacheEntry(context, null, null, tableName, entryId, "Cache", userName, null, "Get", false);
			if (cacheEntry != null) {
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "found Cache entry");
				return cacheEntry;
			}
		}

		return null;
	}

	public List<StatisticsResultInfo> getEntryStatistics(ARPluginContext context, String tableName,
			List<ARVendorField> fields, long transid, QualifierInfo qual, ArithmeticOrRelationalOperand arithOp,
			int statistic, int[] groupBy) throws ARException {
		Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "getEntryStatistics() on " + tableName);
		getConfigDetails(context, false);
		return super.getEntryStatistics(context, tableName, fields, transid, qual, arithOp, statistic, groupBy);
	}

	public List<Entry> getListEntryWithFields(ARPluginContext context, String tableName, List<ARVendorField> fields,
			long transid, QualifierInfo qual, List<SortInfo> sortList, List<EntryListFieldInfo> fieldList, int startAt,
			int maxReturn, OutputInteger numMatches) throws ARException {
		Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "getListEntryWithFields() on " + tableName);
		getConfigDetails(context, false);

		globalElementMap = new TreeMap<String, String>();

		long start = System.currentTimeMillis();

		String userName = Functions.setUserName(context);

		List<Entry> entryList = new ArrayList<Entry>();

		long now = (System.currentTimeMillis() / 1000);

		Map<String, Object> fieldNameValueMap = Functions.parseQual(qual, fields, context);

		if (fieldNameValueMap.containsKey("booleanResult") && fieldNameValueMap.get("booleanResult").equals("false")) {
			return null;
		}

		if (tableName.equals("pluginFunctions")) {
			manageCacheEntry(context, null, null, tableName, null, "ResultList", null, null, "Delete", false);
			Entry entry = pluginFunctions(fieldNameValueMap, context, transid);
			if (entry != null) {
				manageCacheEntry(context, null, null, tableName, entry.getEntryId(), "ResultList", userName, entry,
						"Add", false);
				entryList.add(entry);
			}
			return entryList;
		}

		fieldOptionalList = new ArrayList<String>();
		List<Entry> tableMetaDataList = getTableMetaData(context, tableName, fields, fieldNameValueMap, "GET", false,
				transid);

		long removeBefore = (now - keepResultSeconds);

		// iterating over the results list cache
		List<Entry> resultsEntryList = getListCacheEntry(context, null, null, null, "ResultList", true);

		for (Entry entry : resultsEntryList) {
			long cacheTime = ((Timestamp) entry.get(APLConstants.fidCreateDate).getValue()).getValue();
			if (cacheTime < removeBefore) {
				manageCacheEntry(context, null, null, null, entry.getEntryId(), null, null, null, "Delete", false);
			}
		}

		List<Entry> cacheEntryList = getListCacheEntry(context, null, null, null, "Cache", true);

		Map<String, Long> tableCacheSecondsMap = new HashMap<String, Long>();

		for (Entry entry : cacheEntryList) {
			// iterating over the cached entries
			removeBefore = (now - keepCacheSeconds);

			String curTableName = entry.get(APLConstants.fidCacheTableName).toString();

			long cacheTime = 0;

			if (tableCacheSecondsMap.containsKey(curTableName)) {
				removeBefore = tableCacheSecondsMap.get(curTableName);
			} else {
				try {
					int[] tableFields = { APLConstants.fidTableCacheSeconds };
					List<Entry> tableEntryList = conn.getListEntryObjects(tableForm,
							Functions.getActiveTableQual(curTableName), 0, 0, null, tableFields, false, null);
					if (tableEntryList.size() > 0) {
						Entry tableEntry = tableEntryList.get(0);
						Value tableCacheSecondsValue = tableEntry.get(APLConstants.fidTableCacheSeconds);
						if (tableCacheSecondsValue.getDataType() == DataType.INTEGER) {
							int tableCacheSeconds = tableCacheSecondsValue.getIntValue();
							if (tableCacheSeconds != keepCacheSeconds) {
								removeBefore = (now - tableCacheSeconds);
								tableCacheSecondsMap.put(curTableName, removeBefore);
							} else
								tableCacheSecondsMap.put(curTableName, removeBefore);
						} else
							tableCacheSecondsMap.put(curTableName, removeBefore);
					} else
						manageCacheEntry(context, null, null, null, entry.getEntryId(), null, null, null, "Delete",
								false);
				} catch (ARException e) {
					Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, "Error getting tableCacheSeconds: " + e.getMessage());
				}
			}

			Value cacheSubmitDateTimeValue = entry.get(APLConstants.fidCreateDate);
			if (cacheSubmitDateTimeValue != null && cacheSubmitDateTimeValue.getDataType() == DataType.TIME)
				cacheTime = ((Timestamp) cacheSubmitDateTimeValue.getValue()).getValue();

			if (cacheTime < removeBefore) {
				manageCacheEntry(context, null, null, null, entry.getEntryId(), null, null, null, "Delete", false);
			}
		}

		if (tableMetaDataList != null)
			return tableMetaDataList;

		boolean outputUnusedElements = false;
		boolean bypassCache = false;

		String s = (String)fieldNameValueMap.get("OutputUnusedElements");
		if (s != null && !s.isEmpty())
			outputUnusedElements = true;
		s = (String)fieldNameValueMap.get("BypassCache");
		if (s != null && !s.isEmpty())
			bypassCache = true;

		// ok, we now have a valid 'URL' that needs to be sent, now it's time to
		// use that URL to connect to the service
		long cacheTime = 0;

		String getTemplate = Functions.replaceTemplateValues(tableGetTemplate, fieldNameValueMap, fieldOptionalList,
				aplFieldMap, templateFieldDelimiter);

		cacheEntryList = getListCacheEntry(context, tableName, encodedUrl, getTemplate, "Cache", false);

		// ok now that we have the URL, we are going to check the URL to see if
		// it's in cache at the moment
		if (bypassCache) {
			manageCacheEntry(context, null, null, tableName, null, null, null, null, "Delete", false);
		} else {
			for (Entry entry : cacheEntryList) {
				boolean errorField = false;
				boolean errorOccurred = false;
				int errorFieldID = 0;
				for (ARVendorField field : fields) {
					if (field.getName().equals("ErrorOutput")) {
						errorField = true;
						errorFieldID = field.getFieldId();
						break;
					}
				}

				Entry entryObject = Functions.deserializeEntryObject(context,
						entry.get(APLConstants.fidCacheEntryObject).toString());
				if (errorField) {
					Value errorValue = entryObject.get(errorFieldID);
					if (errorValue != null) {
						String error = errorValue.toString();
						if (!error.isEmpty()) {
							errorOccurred = true;
						}
					}
				}

				if (errorOccurred) {
					cacheTime = 0;
					manageCacheEntry(context, null, null, null, entry.getEntryId(), null, null, null, "Delete", false);
				} else {
					if (debug)
						Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "Using Cached results for entry " + entry.getEntryId());
					entryList.add(entryObject);
				}
			}
		}

		if (entryList.size() == 0) {
			if (debug)
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "Retrieving results from Restful Service");

			List<Entry> restResultsList = makeRestCall(context, tableName, fields, tableGetMethod, getTemplate,
					tableGetResponseCode, fieldNameValueMap, transid);

			if (restResultsList != null)
				return restResultsList;

			InputStream is = null;

			try {
				is = webConn.getInputStream();
			} catch (IOException e) {
				is = ((HttpURLConnection) webConn).getErrorStream();
			}

			String output = Functions.processInputStream(is, Functions.getCharacterSet(webConn.getContentType()));
			if (output.startsWith("<")) {
				is = Functions.stringToIS(output, webConn.getContentType());
				SAXParserFactory factory = SAXParserFactory.newInstance();
				XMLParser handler = new XMLParser(tableRecordIndicatorElement, null, fields);

				try {
					SAXParser parser = factory.newSAXParser();
					parser.parse(is, handler);
				} catch (SAXException e) {
					String message = "SAXException parsing XML: " + e.getMessage();
					Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
					return generateErrorEntryList(message, fields, debug, context, tableName, encodedUrl, transid);
				} catch (ParserConfigurationException e) {
					String message = "ParserConfigurationException parsing XML: " + e.getMessage();
					Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
					return generateErrorEntryList(message, fields, debug, context, tableName, encodedUrl, transid);
				} catch (IOException e) {
					String message = "IOException parsing XML: " + e.getMessage();
					Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
					return generateErrorEntryList(message, fields, debug, context, tableName, encodedUrl, transid);
				}

				globalElementMap = handler.getGlobalElementMap(arrayDelimiter);
				recordElementMap = handler.getRecordElementMap(arrayDelimiter);
				recordElementList = handler.getRecordElementList();
			} else if (output.startsWith("{") || output.startsWith("[")) {
				JSONParser handler = null;
				handler = new JSONParser(fields, tableRecordIndicatorElement, tableRecordIndicatorLevel, skipLevelList);
				handler.parse(output, arrayDelimiter);

				globalElementMap = handler.getGlobalElementMap();
				recordElementMap = handler.getRecordElementMap();
				recordElementList = handler.getRecordElementList();
			} // end if for XML/JSON

			// now it's time to process the headers and add them to the globalElementMap
			if (globalElementMap == null)
				globalElementMap = new TreeMap<String, String>();

			globalElementMap.putAll(Functions.generateHeaderElements(webConn.getHeaderFields(), context, debug));

			((HttpURLConnection) webConn).disconnect();

			int recordCount = 0;

			if (recordElementList.size() > recordCount)
				recordCount = recordElementList.size();
			if (globalElementMap.size() > 0 && recordCount == 0)
				recordCount = 1;

			if (debug) {
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, recordCount + " record(s) in resultset");
				outputElements(context);
			}

			Map<String, String> usedElements = new TreeMap<String, String>();

			if (cacheTime <= 0) {
				cacheTime = now;
			}

			if (recordCount > 0) {
				int i = 1;
				do {
					Entry entry = new Entry();
					String entryID = Functions.generateEntryID(i, transid);
					entry.setEntryId(entryID);

					if (debug)
						Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "Setting EntryID = " + entryID);

					// process the 'local elements'
					for (ARVendorField vendorField : fields) {
						String fieldName = vendorField.getName();

						APLField aplField = null;
						if (aplFieldMap.containsKey(fieldName)) {
							aplField = aplFieldMap.get(fieldName);
						}
						if (aplField == null) {
							aplField = new APLField();
						}

						if (aplField != null && aplField.isLocalElement()) {
							globalElementMap.put(fieldName, (String)fieldNameValueMap.get(fieldName));
							if (debug)
								Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "Adding LocalElement " + fieldName);
						}

					}

					// add response code to the globalElements
					globalElementMap.put("ReturnCode", String.valueOf(responseCode));

					for (ARVendorField vendorField : fields) {
						String fieldName = vendorField.getName();

						APLField aplField = null;
						if (aplFieldMap.containsKey(fieldName)) {
							aplField = aplFieldMap.get(fieldName);
						}
						if (aplField == null) {
							aplField = new APLField();
						}

						int fieldID = vendorField.getFieldId();

						String fieldType = aplField.getType();
						String attFileNameElement = null;
						String attURLElement = null;
						String attSizeElement = null;
						boolean localElement = false;

						if (aplField != null) {
							attFileNameElement = aplField.getAttFileName();
							attURLElement = aplField.getAttURL();
							attSizeElement = aplField.getAttSize();
							fieldType = aplField.getType();
							localElement = aplField.isLocalElement();
						}

						if (localElement) {

						}

						if (fieldType == null)
							fieldType = "";

						if (attFileNameElement == null) {
							attFileNameElement = "";
						}

						if (attURLElement == null) {
							attURLElement = "";
						}

						if (attSizeElement == null) {
							attSizeElement = "";
						}

						String attFileName = "";
						String attURL = "";
						long attSize = 0;

						String returnValue = globalElementMap.get(fieldName);

						if (fieldID == 1) {
							entry.put(fieldID, new Value(entryID, DataType.CHAR));
						}

						Value value = null;

						// processing globalElements returnValues
						if (fieldType.equalsIgnoreCase("ATTACHMENT")) {
							if (!attFileNameElement.isEmpty() && globalElementMap.containsKey(attFileNameElement)) {
								attFileName = globalElementMap.get(attFileNameElement);
							}

							if (!attURLElement.isEmpty() && globalElementMap.containsKey(attURLElement)) {
								attURL = globalElementMap.get(attURLElement);
							}

							if (!attSizeElement.isEmpty() && globalElementMap.containsKey(attSizeElement)) {
								attSize = Long.valueOf(globalElementMap.get(attSizeElement));
							}

							if (attFileName.isEmpty())
								attFileName = "notIdentified";

							if (attFileName.equalsIgnoreCase(attURL)) { // if the filename is the same as the URL, pull
																		// the file name out of the URL
								int fileNameStart = attFileName.lastIndexOf("/");

								if (fileNameStart > 0) {
									attFileName = attFileName.substring(fileNameStart + 1);
								}
							}

							if (!attURL.isEmpty()) {
								// value = new Value(getAttachment(context, attFileName, attURL),
								// DataType.ATTACHMENT);
								AttachmentValue av = new AttachmentValue();
								av.setName(attFileName);
								av.setFilePath(attURL);
								av.setOriginalSize(attSize);
								value = new Value(av, DataType.ATTACHMENT);
							}

							if (debug)
								Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "Adding to " + entryID + ": " + fieldName + ": " + attFileName + " (" + attSize + ")");

							entry.put(fieldID, value);
						} else if (returnValue != null && !returnValue.isEmpty()) {
							if (fieldType.equalsIgnoreCase("INT")) {
								value = new Value(returnValue, DataType.INTEGER);
							} else if (fieldType.equalsIgnoreCase("DATETIME")) {
								value = new Value(returnValue, DataType.TIME);
							} else {
								value = new Value(returnValue, DataType.CHAR);
							}
							if (debug)
								Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "Adding to " + entryID + ": " + fieldName + ": '" + returnValue + "'");

							entry.put(fieldID, value);
						}

						if (globalElementMap.containsKey(fieldName)) {
							usedElements.put(fieldName, returnValue);
						}

						Map<String, String> recordElementMap = null;

						// processing recordElementList
						if (recordElementList.size() >= i) {
							recordElementMap = recordElementList.get(i - 1);
							String recordReturnValue = recordElementMap.get(fieldName);

							if (recordElementMap.containsKey(fieldName)) {
								usedElements.put(fieldName, recordReturnValue);
							}

							value = null;

							if (fieldType.equalsIgnoreCase("ATTACHMENT")) {
								if (!attFileNameElement.isEmpty() && recordElementMap.containsKey(attFileNameElement)) {
									attFileName = recordElementMap.get(attFileNameElement);
								}

								if (!attURLElement.isEmpty() && recordElementMap.containsKey(attURLElement)) {
									attURL = recordElementMap.get(attURLElement);
								}

								if (!attSizeElement.isEmpty() && recordElementMap.containsKey(attSizeElement)) {
									attSize = Long.valueOf(recordElementMap.get(attSizeElement));
								}

								if (attFileName.isEmpty())
									attFileName = "notIdentified";

								if (!attURL.isEmpty()) {
									// value = new Value(getAttachment(context, attFileName, attURL),
									// DataType.ATTACHMENT);
									AttachmentValue av = new AttachmentValue();
									av.setName(attFileName);
									av.setFilePath(attURL);
									av.setOriginalSize(attSize);
									value = new Value(av, DataType.ATTACHMENT);
								}

								if (debug)
									Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "Adding to " + entryID + ": " + fieldName + ": " + attFileName + " (" + attSize + ")");

								entry.put(fieldID, value);
							} else if (recordReturnValue != null && !recordReturnValue.isEmpty()) {
								if (fieldType.equals("INT")) {
									try {
										Integer.parseInt(recordReturnValue);
										value = new Value(recordReturnValue, DataType.INTEGER);
									} catch (NumberFormatException e) {
										value = new Value();
									}
								} else if (fieldType.equals("DATETIME")) {
									try {
										Integer.parseInt(recordReturnValue);
										value = new Value(recordReturnValue, DataType.TIME);
									} catch (NumberFormatException e) {
										value = new Value();
									}
								} else {
									value = new Value(recordReturnValue, DataType.CHAR);
								}

								if (debug)
									Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "Adding to " + entryID + ": " + fieldName + ": '" + recordReturnValue + "'");

								entry.put(fieldID, value);
							}
						}
					}
					entryList.add(entry);
					manageCacheEntry(context, encodedUrl, getTemplate, tableName, entry.getEntryId(), "Cache", userName,
							entry, "Add", false);
					i++;
				} while ((i - 1) < recordElementList.size());
			}

			if (outputUnusedElements) {
				if (globalElementMap != null) {
					for (String element : globalElementMap.keySet()) {
						if (!usedElements.containsKey(element)) {
							Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "Unused Global Element/Value: " + element + ":" + globalElementMap.get(element));
						}
					}
				}
				if (recordElementMap != null) {
					for (String element : recordElementMap.keySet()) {
						if (!usedElements.containsKey(element)) {
							Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "Unused Record Element/Value: " + element + ":" + recordElementMap.get(element));
						}
					}
				}
			}

			if (cacheTime <= 0) {
				cacheTime = now;
			}
		}

		List<Integer> elfiFieldIDs = new ArrayList<Integer>();
		for (EntryListFieldInfo elfi : fieldList) {
			elfiFieldIDs.add(elfi.getFieldId());
		}

		// make a copy of the cache for the results, but give it a unique
		// resultset and ids
		// also, make a copy of the entries returned and restrict them to the
		// columns requested
		resultsEntryList = new ArrayList<Entry>();
		List<Entry> resultsListEntryList = new ArrayList<Entry>();
		int i = 1;
		int i2 = 0;
		for (Entry curEntry : entryList) {
			Entry newResultsListEntry = (Entry) curEntry.clone();
			Entry newEntry = (Entry) curEntry.clone();
			String newEntryID = Functions.generateEntryID(i, transid);
			newResultsListEntry.setEntryId(newEntryID);
			newEntry.setEntryId(newEntryID);
			// loop through the results list and remove any fields (other than
			// 1) that aren't listed in the results list
			Iterator<java.util.Map.Entry<Integer, Value>> resultsListEntryIterator = newResultsListEntry.entrySet()
					.iterator();
			while (resultsListEntryIterator.hasNext()) {
				java.util.Map.Entry<Integer, Value> resultsFieldsEntry = resultsListEntryIterator.next();
				if (!elfiFieldIDs.contains(resultsFieldsEntry.getKey()) && resultsFieldsEntry.getKey() != 1) {
					resultsListEntryIterator.remove();
				}
			}
			// adding local filter capabilities to allow the limiting of what
			// entries are returned not based on the return from the web
			// service, but based on local controls
			if (Functions.processLocalFilter(qual, fields, fieldLocalFilterList, curEntry, context)) {
				if ((i2) >= startAt && (i2) <= maxReturn) {
					resultsListEntryList.add(newResultsListEntry);
					manageCacheEntry(context, null, null, tableName, newEntry.getEntryId(), "ResultList", userName,
							newEntry, "Add", false);
					i2++;
				} else {
					if (debug)
						Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "startAt or maxReturn removed record: " + curEntry.getEntryId());
				}
			} else {
				if (debug)
					Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "Local Filtering removing record: " + curEntry.getEntryId());
			}
			i++;
		}

		long end = System.currentTimeMillis();
		long duration = (end - start);

		if (debug)
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "getListEntryWithFields took " + duration + " ms");

		if (numMatches != null) {
			numMatches.setValue(resultsListEntryList.size());
		}

		return resultsListEntryList;
	}

	public List<VendorForm> getListForms(ARPluginContext context) throws ARException {
		Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "getListForms()");
		getConfigDetails(context, false);

		checkConnection(context);

		ArrayList<VendorForm> list = new ArrayList<VendorForm>();

		List<SortInfo> tableSort = new ArrayList<SortInfo>();
		tableSort.add(new SortInfo(APLConstants.fidTableName, 1));

		int[] tableFields = { APLConstants.fidTableName };

		List<Entry> tableEntryList = conn.getListEntryObjects(tableForm, APLConstants.statusActiveQual, 0, 0, tableSort,
				tableFields, false, null);

		// necessary for creating the 'VendorForm' object
		String pluginName = context.getPluginInfo().getName();

		// loop through the returned data and build a list to display as
		// available 'tables'

		VendorForm form = new VendorForm(pluginName, "pluginFunctions");
		list.add(form);

		for (Entry entry : tableEntryList) {
			form = new VendorForm(pluginName, entry.get(APLConstants.fidTableName).toString());
			list.add(form);
		}

		return list;
	}

	public List<ARVendorField> getMultipleFields(ARPluginContext context, VendorForm form) throws ARException {
		Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "getMultipleFields() on " + form.getName());
		getConfigDetails(context, false);

		ArrayList<ARVendorField> fieldList = new ArrayList<ARVendorField>();

		String tableName = form.getTableName();

		if (tableName.equals("pluginFunctions")) {
			ARVendorField request = new ARVendorField("Request ID", 0, Constants.AR_DATA_TYPE_CHAR,
					new CharacterFieldLimit(0, Constants.AR_MENU_APPEND, Constants.AR_QBE_MATCH_EQUAL, "", "", 0));
			ARVendorField function = new ARVendorField("Function", 0, Constants.AR_DATA_TYPE_CHAR,
					new CharacterFieldLimit(0, Constants.AR_MENU_APPEND, Constants.AR_QBE_MATCH_EQUAL, "", "", 0));
			ARVendorField output = new ARVendorField("Output", 0, Constants.AR_DATA_TYPE_CHAR,
					new CharacterFieldLimit(0, Constants.AR_MENU_APPEND, Constants.AR_QBE_MATCH_EQUAL, "", "", 0));
			ARVendorField input1 = new ARVendorField("Input1", 0, Constants.AR_DATA_TYPE_CHAR,
					new CharacterFieldLimit(0, Constants.AR_MENU_APPEND, Constants.AR_QBE_MATCH_EQUAL, "", "", 0));
			ARVendorField input2 = new ARVendorField("Input2", 0, Constants.AR_DATA_TYPE_CHAR,
					new CharacterFieldLimit(0, Constants.AR_MENU_APPEND, Constants.AR_QBE_MATCH_EQUAL, "", "", 0));
			ARVendorField input3 = new ARVendorField("Input3", 0, Constants.AR_DATA_TYPE_CHAR,
					new CharacterFieldLimit(0, Constants.AR_MENU_APPEND, Constants.AR_QBE_MATCH_EQUAL, "", "", 0));
			ARVendorField input4 = new ARVendorField("Input4", 0, Constants.AR_DATA_TYPE_CHAR,
					new CharacterFieldLimit(0, Constants.AR_MENU_APPEND, Constants.AR_QBE_MATCH_EQUAL, "", "", 0));
			ARVendorField input5 = new ARVendorField("Input5", 0, Constants.AR_DATA_TYPE_CHAR,
					new CharacterFieldLimit(0, Constants.AR_MENU_APPEND, Constants.AR_QBE_MATCH_EQUAL, "", "", 0));
			ARVendorField input6 = new ARVendorField("Input6", 0, Constants.AR_DATA_TYPE_CHAR,
					new CharacterFieldLimit(0, Constants.AR_MENU_APPEND, Constants.AR_QBE_MATCH_EQUAL, "", "", 0));
			ARVendorField input7 = new ARVendorField("Input7", 0, Constants.AR_DATA_TYPE_CHAR,
					new CharacterFieldLimit(0, Constants.AR_MENU_APPEND, Constants.AR_QBE_MATCH_EQUAL, "", "", 0));
			ARVendorField debug = new ARVendorField("Debug", 0, Constants.AR_DATA_TYPE_CHAR,
					new CharacterFieldLimit(0, Constants.AR_MENU_APPEND, Constants.AR_QBE_MATCH_EQUAL, "", "", 0));
			fieldList.add(request);
			fieldList.add(function);
			fieldList.add(output);
			fieldList.add(input1);
			fieldList.add(input2);
			fieldList.add(input3);
			fieldList.add(input4);
			fieldList.add(input5);
			fieldList.add(input6);
			fieldList.add(input7);
			fieldList.add(debug);
		} else {
			ARVendorField entryField = new ARVendorField("EntryID", 0, Constants.AR_DATA_TYPE_CHAR,
					new CharacterFieldLimit(15, Constants.AR_MENU_APPEND, Constants.AR_QBE_MATCH_EQUAL, "", "", 0));
			ARVendorField outputUnusedElementsField = new ARVendorField("OutputUnusedElements", 0,
					Constants.AR_DATA_TYPE_CHAR,
					new CharacterFieldLimit(30, Constants.AR_MENU_APPEND, Constants.AR_QBE_MATCH_EQUAL, "", "", 0));
			ARVendorField bypassCacheField = new ARVendorField("BypassCache", 0, Constants.AR_DATA_TYPE_CHAR,
					new CharacterFieldLimit(30, Constants.AR_MENU_APPEND, Constants.AR_QBE_MATCH_EQUAL, "", "", 0));
			ARVendorField debugField = new ARVendorField("Debug", 0, Constants.AR_DATA_TYPE_CHAR,
					new CharacterFieldLimit(30, Constants.AR_MENU_APPEND, Constants.AR_QBE_MATCH_EQUAL, "", "", 0));
			ARVendorField errorOutput = new ARVendorField("ErrorOutput", 0, Constants.AR_DATA_TYPE_CHAR,
					new CharacterFieldLimit(30, Constants.AR_MENU_APPEND, Constants.AR_QBE_MATCH_EQUAL, "", "", 0));

			ARVendorField returnCode = new ARVendorField("ReturnCode", 0, Constants.AR_DATA_TYPE_INTEGER,
					new IntegerFieldLimit(0, 1000));

			fieldList.add(entryField);
			fieldList.add(outputUnusedElementsField);
			fieldList.add(bypassCacheField);
			fieldList.add(debugField);
			fieldList.add(errorOutput);
			fieldList.add(returnCode);

			checkConnection(context);

			int[] tableFields = { APLConstants.fidGuid };

			List<Entry> tableEntryList = conn.getListEntryObjects(tableForm, Functions.getActiveTableQual(tableName), 0,
					0, null, tableFields, false, null);

			// there should be only one return for the table, so we won't loop
			// through, we will make some assumptions
			if (tableEntryList != null && tableEntryList.size() == 1) {
				String tableGUID = tableEntryList.get(0).get(APLConstants.fidGuid).toString();

				// now that we have the table GUID, we need to go get the field data
				List<SortInfo> fieldSort = new ArrayList<SortInfo>();
				fieldSort.add(new SortInfo(APLConstants.fidFieldName, 1)); // sort by field name

				int[] fieldFields = { APLConstants.fidFieldName, APLConstants.fidFieldType };

				List<Entry> fieldEntryList = conn.getListEntryObjects(fieldForm, Functions.getActiveGUIDQual(tableGUID),
						0, 0, fieldSort, fieldFields, false, null);

				for (Entry entry : fieldEntryList) {
					String fieldName = entry.get(APLConstants.fidFieldName).toString();
					String fieldType = entry.get(APLConstants.fidFieldType).toString();

					if (fieldType == null) {
						fieldType = "";
					}

					int type = Constants.AR_DATA_TYPE_CHAR;
					FieldLimit limit = null;

					if (fieldType.equalsIgnoreCase("INT")) {
						type = Constants.AR_DATA_TYPE_INTEGER;
						limit = new IntegerFieldLimit(-2147483647, 2147483647);
					} else if (fieldType.equalsIgnoreCase("DATETIME")) {
						type = Constants.AR_DATA_TYPE_TIME;
						limit = new DateTimeFieldLimit();
					} else if (fieldType.equalsIgnoreCase("ATTACHMENT")) {
						type = Constants.AR_DATA_TYPE_ATTACH;
						limit = new AttachmentFieldLimit();
					} else { // else is assuming Char
						limit = new CharacterFieldLimit(0, Constants.AR_MENU_APPEND, Constants.AR_QBE_MATCH_EQUAL, "",
								"", 0);
					}

					ARVendorField field = new ARVendorField(fieldName, 0, type, limit);

					fieldList.add(field);
				}
			} else {
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, "Unable to find record on " + tableForm + " for record '" + tableName + "'");
			}
		}

		return fieldList;
	}

	public void rollback(ARPluginContext context, long transid) throws ARException {
		Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "rollBack(): " + transid);
	}

	public void setBLOB(ARPluginContext context, String tableName, List<ARVendorField> fields, long transid,
			String entryId, int fieldId, AttachmentValue content) throws ARException {
		Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "setBlob() on " + tableName);
		getConfigDetails(context, false);

		int[] fieldIDArray = new int[fields.size()];

		for (int i = 0; i < fields.size(); i++) {
			int fieldID = fields.get(i).getFieldId();
			fieldIDArray[i] = fieldID;
		}

		Entry entry = getEntry(context, tableName, fields, transid, entryId, fieldIDArray);

		Value value = new Value(content);

		entry.put(fieldId, value);

		setEntry(context, tableName, fields, transid, entryId, entry, new Timestamp());
	}

	public void setEntry(ARPluginContext context, String tableName, List<ARVendorField> fields, long transid,
			String entryId, Entry entry, Timestamp lastModified) throws ARException {
		Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "setEntry on " + tableName + " on " + entry.getEntryId());
		getConfigDetails(context, false);

		globalElementMap = new TreeMap<String, String>();

		debug = Functions.getDebug(fields, entry, alwaysLogDetails);

		String userName = Functions.setUserName(context);

		// build field array
		int[] fieldIDArray = new int[fields.size()];
		for (int i = 0; i < fields.size(); i++) {
			int fieldID = fields.get(i).getFieldId();
			fieldIDArray[i] = fieldID;
		}

		// get the current entry from the process
		Entry currentEntry = getEntry(context, tableName, fields, transid, entryId, fieldIDArray);

		// get the current and changed entries into maps
		Map<String, Object> fieldNameValueMap = Functions.processEntry(currentEntry, fields, aplFieldMap);

		Map<String, Object> fieldNameValueMapChanging = Functions.processEntry(entry, fields, aplFieldMap);

		// overlay the changes onto the current
		for (String key : fieldNameValueMapChanging.keySet()) {
			fieldNameValueMap.put(key, fieldNameValueMapChanging.get(key));
		}

		// load Metadata
		fieldOptionalList = new ArrayList<String>();
		List<Entry> tableMetaDataList = getTableMetaData(context, tableName, fields, fieldNameValueMap, "SET", false,
				transid);

		if (tableMetaDataList != null) {
			String message = Functions.getErrorString(fields, tableMetaDataList);
			throw new ARException(Constants.AR_RETURN_ERROR, ARErrors.AR_PLUGIN_ERROR_UNKOWN, message);
		}

		// replace values in templates
		String setTemplate = Functions.replaceTemplateValues(tableSetTemplate, fieldNameValueMap, fieldOptionalList,
				aplFieldMap, templateFieldDelimiter);

		// make the restful call as defined
		List<Entry> restResultsList = makeRestCall(context, tableName, fields, tableSetMethod, setTemplate,
				tableSetResponseCode, fieldNameValueMap, transid);

		if (restResultsList != null) {
			String message = Functions.getErrorString(fields, restResultsList);
			throw new ARException(Constants.AR_RETURN_ERROR, ARErrors.AR_PLUGIN_ERROR_UNKOWN, message);
		}

		((HttpURLConnection) webConn).disconnect();

		// get current cache entry
		Entry cacheEntry = manageCacheEntry(context, null, null, tableName, entryId, "ResultList", userName, null,
				"Get", true);
		Entry cacheEntryObject = Functions.deserializeEntryObject(context,
				cacheEntry.get(APLConstants.fidCacheEntryObject).toString());

		// loop through the updated entry and update the entry values
		for (Integer fieldID : entry.keySet()) {
			cacheEntryObject.put(fieldID, entry.get(fieldID));
		}

		try {
			cacheEntry.put(APLConstants.fidCacheEntryObject,
					new Value(Functions.serializeEntryObject(cacheEntryObject)));
		} catch (IOException e) {
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, "Error serializing entry object on setEntry:" + e.getMessage());
		}

		conn.setEntry(cacheForm, cacheEntry.getEntryId(), cacheEntry, new Timestamp(),
				Constants.AR_JOIN_SETOPTION_NONE);
	}

	public void initialize(ARPluginContext context) throws ARException {
		System.out.println("Initializing plug-in: APLRestfulARDBCPlugin version " + version);

		CookieHandler.setDefault(new CookieManager(null, CookiePolicy.ACCEPT_ALL));

		getConfigDetails(context, true);
	}

	private void getConfigDetails(ARPluginContext context, boolean log) {
		alwaysLogDetails = false;
		arrayDelimiter = ",";
		cacheForm = "APL:RestfulVendorForms:Cache";
		configForm = "APL:RestfulVendorForms:Config";
		fieldForm = "APL:RestfulVendorForms:Fields";
		functionForm = "APL:RestfulVendorForms:Functions";
		keepCacheSeconds = 86400;
		keepResultSeconds = 300;
		keyStoreAlgorithm = "PKIX";
		keyStoreFile = "";
		keyStoreFormat = "PKCS12";
		keyStorePass = "";
		keyStoreProtocol = "TLS";
		proxyPass = "";
		proxyPort = 0;
		proxyServer = "";
		proxyUser = "";
		remedyServerName = "localhost";
		remedyServerPort = 0;
		tableForm = "APL:RestfulVendorForms:Table";
		templateFieldDelimiter = "|";
		timeout = 90;
		proxy = null;

		if (log)
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "Initializing plug-in: APLRestfulARDBCPlugin version " + version);

		Map<String, String> configMap = new HashMap<String, String>();

		String configServerName = context.getARConfigEntry("Server-Connect-Name");
		if (configServerName != null && !configServerName.isEmpty())
			remedyServerName = configServerName;
		else {
			configServerName = context.getARConfigEntry("Server-Name");
			if (configServerName != null && !configServerName.isEmpty())
				remedyServerName = configServerName;
		}
		configServerName = processConfig(context, "RemedyServerName", configMap);
		if (configServerName != null)
			remedyServerName = configServerName;

		String configServerPort = context.getARConfigEntry("TCD-Specific-Port");
		Integer tcdPort = null;
		if (configServerPort != null && !configServerPort.isEmpty()) {
			try {
				tcdPort = Integer.parseInt(configServerPort);
				remedyServerPort = tcdPort;
			} catch (Exception e) {
				if (log)
					Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, "Error setting TCP Port from ar.cfg setting TCD-Specific-Port: " + e.getMessage());
			}
		}
		configServerPort = processConfig(context, "RemedyServerPort", configMap);
		if (configServerPort != null)
			try {
				tcdPort = Integer.parseInt(configServerPort);
				remedyServerPort = tcdPort;
			} catch (Exception e) {
			}

		// now that I have a server name and port, go get the config from data
		checkConnection(context);
		String configConfigForm = processConfig(context, "ConfigForm", configMap);
		if (configConfigForm != null)
			configForm = configConfigForm;
		else {
			configConfigForm = Functions.getFormName(context, conn, APLConstants.configFields, configForm);
			if (configConfigForm != null)
				configForm = configConfigForm;
		}
		if (log)
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "ConfigForm used will be " + configForm);

		List<Entry> configEntryList = null;
		try {
			configEntryList = conn.getListEntryObjects(configForm, APLConstants.statusActiveQual, 0, 0, null,
					APLConstants.configFields, false, null);
			for (Entry entry : configEntryList) {
				Value entryKey = entry.get(APLConstants.fidConfigName);
				Value entryValue = entry.get(APLConstants.fidConfigValue);
				if (entryKey != null && entryKey.getDataType() != DataType.NULL && entryValue != null
						&& entryValue.getDataType() != DataType.NULL) {
					configMap.put(entryKey.toString(), entryValue.toString());
				}
			}
		} catch (ARException e) {
		}

		configServerName = processConfig(context, "RemedyServerName", configMap);
		if (configServerName != null)
			remedyServerName = configServerName;
		configServerPort = processConfig(context, "RemedyServerPort", configMap);
		if (configServerPort != null) {
			try {
				tcdPort = Integer.parseInt(configServerPort);
				remedyServerPort = tcdPort;
			} catch (Exception e) {
				if (log)
					Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, "Error setting TCP Port: " + e.getMessage());
			}

		}

		String serverOut = "Remedy server name being used is '" + remedyServerName + "'";
		if (remedyServerPort > 0)
			serverOut += " on port " + remedyServerPort;

		if (log)
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, serverOut);

		String configTableForm = processConfig(context, "TableForm", configMap);
		if (configTableForm != null)
			tableForm = configTableForm;
		else {
			configTableForm = Functions.getFormName(context, conn, APLConstants.tableFields, tableForm);
			if (configTableForm != null)
				tableForm = configTableForm;
		}
		if (log)
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "TableForm used will be " + tableForm);

		String configFieldForm = processConfig(context, "FieldForm", configMap);
		if (configFieldForm != null)
			fieldForm = configFieldForm;
		else {
			configFieldForm = Functions.getFormName(context, conn, APLConstants.fieldFields, fieldForm);
			if (configFieldForm != null)
				fieldForm = configFieldForm;
		}
		if (log)
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "FieldForm used will be " + fieldForm);

		String configCacheForm = processConfig(context, "CacheForm", configMap);
		if (configCacheForm != null)
			cacheForm = configCacheForm;
		else {
			configCacheForm = Functions.getFormName(context, conn, APLConstants.cacheFields, cacheForm);
			if (configCacheForm != null)
				cacheForm = configCacheForm;
		}
		if (log)
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO,
					"CacheForm used will be " + cacheForm);

		String configFunctionForm = processConfig(context, "FunctionForm", configMap);
		if (configFunctionForm != null)
			functionForm = configFunctionForm;
		else {
			configFunctionForm = Functions.getFormName(context, conn, APLConstants.functionFields, functionForm);
			if (configFunctionForm != null)
				functionForm = configFunctionForm;
		}
		if (log)
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO,
					"FunctionForm used will be " + functionForm);

		String configKeepCacheSeconds = processConfig(context, "KeepCacheSeconds", configMap);
		if (configKeepCacheSeconds != null) {
			try {
				keepCacheSeconds = Integer.parseInt(configKeepCacheSeconds);
			} catch (Exception e) {
			}
		}
		if (log)
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO,
					"Cache will be stored for " + keepCacheSeconds + " seconds");

		String configKeepResultSeconds = processConfig(context, "KeepResultSeconds", configMap);
		if (configKeepResultSeconds != null) {
			try {
				keepResultSeconds = Integer.parseInt(configKeepResultSeconds);
			} catch (Exception e) {
			}
		}
		if (log)
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO,
					"Results will be stored for " + keepResultSeconds + " seconds");

		String configAlwaysLogDetails = processConfig(context, "AlwaysLogDetails", configMap);
		if (configAlwaysLogDetails != null) {
			alwaysLogDetails = Boolean.parseBoolean(configAlwaysLogDetails);
		}
		if (log)
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO,
					"AlwaysLogDetails is set to " + alwaysLogDetails);

		String configTimeout = processConfig(context, "Timeout", configMap);
		if (configTimeout != null) {
			try {
				timeout = Integer.parseInt(configTimeout);
			} catch (Exception e) {
			}
		}
		if (log)
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO,
					"Timeout for calls will be " + timeout + " seconds");

		String configProxyServer = processConfig(context, "ProxyServer", configMap);
		String configProxyPort = processConfig(context, "ProxyPort", configMap);
		String configProxyUser = processConfig(context, "ProxyUser", configMap);
		String configProxyPass = processConfig(context, "ProxyPass", configMap);

		if (configProxyServer != null && configProxyPort != null && !configProxyServer.isEmpty()
				&& !configProxyPort.isEmpty()) {
			try {
				proxyServer = configProxyServer;
				proxyPort = Integer.parseInt(configProxyPort);

				proxy = new Proxy(Proxy.Type.HTTP, new InetSocketAddress(proxyServer, proxyPort));

				String proxyOut = "Proxy server being used is '" + proxyServer + "' on port " + proxyPort;

				if (configProxyUser != null && configProxyPass != null && !configProxyUser.isEmpty()
						&& !configProxyPass.isEmpty()) {
					proxyUser = configProxyUser;
					proxyPass = configProxyPass;

					proxyOut += " with username of " + proxyUser;
				}

				if (log)
					Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, proxyOut);

			} catch (Exception e) {
				if (log)
					Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR,
							"Error setting Proxy configurations, please verify all information is correct: "
									+ e.getMessage());
			}
		}

		String configKeyStoreAlgorithm = processConfig(context, "KeyStoreAlgorithm", configMap);
		if (configKeyStoreAlgorithm != null)
			keyStoreAlgorithm = configKeyStoreAlgorithm;
		if (log)
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO,
					"KeyStoreAlorithm is " + keyStoreAlgorithm);

		String configKeyStoreFormat = processConfig(context, "KeyStoreFormat", configMap);
		if (configKeyStoreAlgorithm != null)
			keyStoreFormat = configKeyStoreFormat;
		if (log)
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO,
					"KeyStoreFormat is " + keyStoreFormat);

		String configKeyStoreFile = processConfig(context, "KeyStoreFile", configMap);
		if (configKeyStoreFile != null)
			keyStoreFile = configKeyStoreFile;
		if (log)
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO,
					"KeyStoreFile is " + keyStoreFile);

		String configKeyStorePass = processConfig(context, "KeyStorePass", configMap);
		if (configKeyStorePass != null)
			keyStorePass = configKeyStorePass;

		String configKeyStoreProtocol = processConfig(context, "KeyStoreProtocol", configMap);
		if (configKeyStoreProtocol != null)
			keyStoreProtocol = configKeyStoreProtocol;
		if (log)
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO,
					"KeyStoreProtocol is " + keyStoreProtocol);

		String configTemplateFieldDelimiter = processConfig(context, "TemplateFieldDelimiter", configMap);
		if (configTemplateFieldDelimiter != null)
			templateFieldDelimiter = configTemplateFieldDelimiter;
		if (log)
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO,
					"TemplateFieldDelimiter is " + templateFieldDelimiter);

		String configArrayDelimiter = processConfig(context, "ArrayDelimiter", configMap);
		if (configArrayDelimiter != null)
			arrayDelimiter = configArrayDelimiter;
		if (log)
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO,
					"ArrayDelimiter is " + arrayDelimiter);
	}

	private String processConfig(ARPluginContext context, String setting, Map<String, String> configMap) {
		String value = null;

		if (configMap.containsKey(setting)) {
			value = configMap.get(setting);
		} else {
			String configValue = context.getConfigItem(setting);
			if (configValue != null && !configValue.isEmpty()) {
				value = configValue;
			}
		}

		return value;
	}

	public void onEvent(ARPluginContext context, int eventId) throws ARException {
		Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "onEvent(): " + eventId);
		super.onEvent(context, eventId);

		initialize(context);
	}

	public void terminate(ARPluginContext context) throws ARException {
		Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "terminate()");

		// when the plugin server tells you to shut down...here it is.
	}

	private void outputElements(ARPluginContext context) {
		Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "Global Elements:");
		for (String key : globalElementMap.keySet()) {
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO,
					key + "|" + globalElementMap.get(key));
		}

		Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "Record Elements:");
		for (Map<String, String> recordElementsMap : recordElementList) {
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "Start Record:");
			for (String key : recordElementsMap.keySet()) {
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO,
						key + "|" + recordElementsMap.get(key));
			}
		}
	}

	private List<Entry> generateErrorEntryList(String message, List<ARVendorField> fields, boolean debug,
			ARPluginContext context, String tableName, String encodedUrl, long transid) throws ARException {
		List<Entry> errorEntryResultsList = new ArrayList<Entry>();
		List<String> errorEntryIDResultsList = new ArrayList<String>();
		boolean addEntry = false;

		String userName = Functions.setUserName(context);

		Entry entry = new Entry();
		if (fields != null) {
			for (ARVendorField field : fields) {
				if (field.getName().equals("ErrorOutput")) {
					String entryID = Functions.generateEntryID(1, transid);
					entry.setEntryId(entryID);
					entry.put(1, new Value(entryID, DataType.CHAR));
					entry.put(field.getFieldId(), new Value(message, DataType.CHAR));
					errorEntryIDResultsList.add(entry.getEntryId());
					addEntry = true;

				} else if (field.getName().equals("ReturnCode")) {
					entry.put(field.getFieldId(), new Value(responseCode, DataType.INTEGER));
				}
			}
			if (tableName != null) {
				manageCacheEntry(context, encodedUrl, null, tableName, entry.getEntryId(), "Cache", userName,
						entry, "Add", false);
			}
		}

		if (errorEntryIDResultsList.size() == 0)
			throw new ARException(Constants.AR_RETURN_ERROR, ARErrors.AR_PLUGIN_ERROR_UNKOWN, message);

		if (addEntry) {
			errorEntryResultsList.add(entry);
		}

		return errorEntryResultsList;
	}

	private List<Entry> getTableMetaData(ARPluginContext context, String tableName, List<ARVendorField> fields,
			Map<String, Object> fieldNameValueMap, String operation, boolean autoGenerate, long transid)
			throws ARException {

		checkConnection(context);

		if (conn == null) {
			String message = "Unable to connect back into Remedy server to get table MetaData";
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
			return generateErrorEntryList(message, fields, false, context, tableName, "", transid);
		}

		// this is where we translate the Remedy query into the restful query,
		// parse the results, and return them as a 'results list'
		List<Entry> tableEntry = conn.getListEntryObjects(tableForm, Functions.getActiveTableQual(tableName), 0, 0,
				null, APLConstants.tableFields, false, null);

		if (tableEntry.size() == 1) {
			tableRequestID = tableEntry.get(0).get(APLConstants.fidTableRequestID).toString();
			tableGUID = tableEntry.get(0).get(APLConstants.fidGuid).toString();
			tableURL = tableEntry.get(0).get(APLConstants.fidTableURL).toString();
			tableURLSuffix = tableEntry.get(0).get(APLConstants.fidTableURLSuffix).toString();
			tableRecordIndicatorElement = tableEntry.get(0).get(APLConstants.fidTableRecordIndicatorElement).toString();
			tableBasicUser = tableEntry.get(0).get(APLConstants.fidTableBasicUser).toString();
			tableRemedyServer = tableEntry.get(0).get(APLConstants.fidTableRemedyServer).toString();
			if (tableRemedyServer != null && !tableRemedyServer.isEmpty())
				remedyServer = true;
			tableCustomHeader = tableEntry.get(0).get(APLConstants.fidTableCustomHeader).toString();
			tableTimeoutS = tableEntry.get(0).get(APLConstants.fidTableTimeout).toString();
			tableGetMethod = tableEntry.get(0).get(APLConstants.fidTableGetMethod).toString();
			tableGetTemplate = tableEntry.get(0).get(APLConstants.fidTableGetTemplate).toString();
			tableCreateMethod = tableEntry.get(0).get(APLConstants.fidTableCreateMethod).toString();
			tableCreateTemplate = tableEntry.get(0).get(APLConstants.fidTableCreateTemplate).toString();
			tableSetMethod = tableEntry.get(0).get(APLConstants.fidTableSetMethod).toString();
			tableSetTemplate = tableEntry.get(0).get(APLConstants.fidTableSetTemplate).toString();
			tableDeleteMethod = tableEntry.get(0).get(APLConstants.fidTableDeleteMethod).toString();
			tableDeleteTemplate = tableEntry.get(0).get(APLConstants.fidTableDeleteTemplate).toString();
			tableCreateRequestIDElement = tableEntry.get(0).get(APLConstants.fidTableCreateRequestIDElement).toString();
			tableBypassSSLChecks = tableEntry.get(0).get(APLConstants.fidTableBypassSSLChecks).toString();
			tableBypassProxy = tableEntry.get(0).get(APLConstants.fidTableBypassProxy).toString();

			if (tableBypassSSLChecks != null && !tableBypassSSLChecks.isEmpty())
				bypassCertCheck = true;
			else
				bypassCertCheck = false;

			if (tableBypassProxy != null && !tableBypassProxy.isEmpty())
				bypassProxy = true;
			else
				bypassProxy = false;

			try {
				tableGetResponseCode = tableEntry.get(0).get(APLConstants.fidTableGetResponseCode).getIntValue();
			} catch (NullPointerException e) {
				tableGetResponseCode = HttpURLConnection.HTTP_OK;
			}

			try {
				tableCreateResponseCode = tableEntry.get(0).get(APLConstants.fidTableCreateResponseCode).getIntValue();
			} catch (NullPointerException e) {
				tableCreateResponseCode = HttpURLConnection.HTTP_CREATED;
			}

			try {
				tableSetResponseCode = tableEntry.get(0).get(APLConstants.fidTableSetResponseCode).getIntValue();
			} catch (NullPointerException e) {
				tableSetResponseCode = HttpURLConnection.HTTP_OK;
			}

			try {
				tableDeleteResponseCode = tableEntry.get(0).get(APLConstants.fidTableDeleteResponseCode).getIntValue();
			} catch (NullPointerException e) {
				tableDeleteResponseCode = HttpURLConnection.HTTP_OK;
			}

			allowAllReturnCodesS = tableEntry.get(0).get(APLConstants.fidTableAllowAllReturnCodes).toString();
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "AllowAll: " + allowAllReturnCodesS);
			if (allowAllReturnCodesS != null && !allowAllReturnCodesS.isEmpty())
				allowAllReturnCodes = true;
			else
				allowAllReturnCodes = false;

			if (tableGetMethod == null || tableGetMethod.isEmpty()) {
				tableGetMethod = "GET";
			}
			if (tableCreateMethod == null || tableCreateMethod.isEmpty()) {
				tableCreateMethod = "POST";
			}
			if (tableSetMethod == null || tableSetMethod.isEmpty()) {
				tableSetMethod = "PUT";
			}
			if (tableDeleteMethod == null || tableDeleteMethod.isEmpty()) {
				tableDeleteMethod = "DELETE";
			}

			try {
				tableRecordIndicatorLevel = tableEntry.get(0).get(APLConstants.fidTableRecordIndicatorLevel)
						.getIntValue();
			} catch (NullPointerException e) {
			}

			try {
				tableSkipElementLevels = tableEntry.get(0).get(APLConstants.fidTableSkipElementLevels).toString();
				String[] tokens = tableSkipElementLevels.split(",");
				for (String token : tokens) {
					try {
						skipLevelList.add(Integer.valueOf(token.trim()));
					} catch (NumberFormatException e) {
						String message = "Error while processing the skipLevelList, " + token
								+ " could not be converted to a number";
						Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
					}
				}
			} catch (NullPointerException e) {
			}
		} else {
			String message = "Query from " + tableForm + " for table '" + tableName + "' returned " + tableEntry.size()
					+ " instead of 1 record.  Please correct that and retry the operation";
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
			return generateErrorEntryList(message, fields, false, context, tableName, "", transid);
		}

		if (tableBasicUser != null) {
			Entry entry = new Entry();
			entry.put(APLConstants.fidTablezFunction, new Value("Decrypt"));
			int[] returnFields = { APLConstants.fidTableBasicPass };
			entry = conn.executeService(tableForm, tableRequestID, entry, returnFields);
			Value tableBasicPassValue = entry.get(APLConstants.fidTableBasicPass);
			if (tableBasicPassValue != null && tableBasicPassValue.getDataType() != DataType.NULL) {
				tableBasicPass = tableBasicPassValue.toString();
			} else {
				tableBasicPass = "";
			}
		} else {
			tableBasicUser = "";
		}

		if (tableURLSuffix == null)
			tableURLSuffix = "";

		if (tableRecordIndicatorElement == null)
			tableRecordIndicatorElement = "";

		if (!autoGenerate)
			urlS = tableURL;

		// now that we have the table GUID, we need to go get the field data
		List<SortInfo> fieldSort = new ArrayList<SortInfo>();
		fieldSort.add(new SortInfo(APLConstants.fidFieldSort, 1)); // sort by field order

		List<Entry> fieldEntryList = conn.getListEntryObjects(fieldForm, Functions.getActiveGUIDQual(tableGUID), 0, 0,
				fieldSort, APLConstants.fieldFields, false, null);

		aplFieldMap = new HashMap<String, APLField>();
		//System.out.println("GetTableMetaData: " + fieldEntryList.size());
		
		for (Entry entry : fieldEntryList) {
			String fieldName = entry.get(APLConstants.fidFieldName).toString();
			String fieldType = entry.get(APLConstants.fidFieldType).toString();
			String fieldPrefix = entry.get(APLConstants.fidFieldPrefix).toString();
			String fieldSuffix = entry.get(APLConstants.fidFieldSuffix).toString();
			String fieldOptional = entry.get(APLConstants.fidFieldOptional).toString();
			String fieldGetURLAppend = entry.get(APLConstants.fidFieldGetURLAppend).toString();
			String fieldCreateURLAppend = entry.get(APLConstants.fidFieldCreateURLAppend).toString();
			String fieldSetURLAppend = entry.get(APLConstants.fidFieldSetURLAppend).toString();
			String fieldDeleteURLAppend = entry.get(APLConstants.fidFieldDeleteURLAppend).toString();
			String fieldLocalFilter = entry.get(APLConstants.fidFieldLocalFilter).toString();
			String fieldAttName = entry.get(APLConstants.fidFieldAttName).toString();
			String fieldAttURL = entry.get(APLConstants.fidFieldAttURL).toString();
			String fieldAttSize = entry.get(APLConstants.fidFieldAttSize).toString();
			String fieldBypassEncoding = entry.get(APLConstants.fidFieldBypassEncoding).toString();
			String fieldLocalElement = entry.get(APLConstants.fidFieldLocalElement).toString();

			//System.out.println(fieldName + ": " + fieldBypassEncoding);
			
			APLField aplField = new APLField();

			aplField.setFieldName(fieldName);
			aplField.setType(fieldType);
			aplField.setAttFileName(fieldAttName);
			aplField.setAttURL(fieldAttURL);
			aplField.setAttSize(fieldAttSize);
			aplField.setBypassEncoding(fieldBypassEncoding);
			aplField.setLocalElement(fieldLocalElement);

			Object fieldValue = null;

			if (fieldNameValueMap != null)
				fieldValue = fieldNameValueMap.get(fieldName);

			boolean optionalInput = false;

			if (fieldOptional != null && fieldOptional.equals("0")) {
				optionalInput = true;
				fieldOptionalList.add(fieldName);
			}

			if (fieldLocalFilter != null && fieldLocalFilter.equals("0")) {
				fieldLocalFilterList.add(fieldName);
			}

			if (fieldPrefix == null)
				fieldPrefix = "";

			if (fieldValue == null)
				fieldValue = "";

			if (fieldSuffix == null)
				fieldSuffix = "";

			if (operation == null)
				operation = "";

			if (!autoGenerate) {
				if (operation.equalsIgnoreCase("GET")) {
					if (fieldGetURLAppend != null) {
						if (!(((String)fieldValue).isEmpty() && optionalInput)) {
							urlS += fieldPrefix;
							urlS += fieldValue;
							urlS += fieldSuffix;
						}
					}
				} else if (operation.equalsIgnoreCase("CREATE")) {
					if (fieldCreateURLAppend != null) {
						if (!(((String)fieldValue).isEmpty() && optionalInput)) {
							urlS += fieldPrefix;
							urlS += fieldValue;
							urlS += fieldSuffix;
						}
					}
				} else if (operation.equalsIgnoreCase("SET")) {
					if (fieldSetURLAppend != null) {
						if (!(((String)fieldValue).isEmpty() && optionalInput)) {
							urlS += fieldPrefix;
							urlS += fieldValue;
							urlS += fieldSuffix;
						}
					}
				} else if (operation.equalsIgnoreCase("DELETE")) {
					if (fieldDeleteURLAppend != null) {
						if (!(((String)fieldValue).isEmpty() && optionalInput)) {
							urlS += fieldPrefix;
							urlS += fieldValue;
							urlS += fieldSuffix;
						}
					}
				}
			}

			aplFieldMap.put(fieldName, aplField);
		}

		APLField debugField = new APLField();
		debugField.setFieldName("Debug");
		debugField.setLocalElement("true");
		aplFieldMap.put("Debug", debugField);

		APLField outputUnusedElementsField = new APLField();
		outputUnusedElementsField.setFieldName("OutputUnusedElements");
		outputUnusedElementsField.setLocalElement("true");
		aplFieldMap.put("OutputUnusedElements", outputUnusedElementsField);

		APLField bypassCacheField = new APLField();
		bypassCacheField.setFieldName("BypassCache");
		bypassCacheField.setLocalElement("true");
		aplFieldMap.put("BypassCache", bypassCacheField);

		if (!autoGenerate)
			urlS += tableURLSuffix;

		debug = Functions.getDebug(fieldNameValueMap, alwaysLogDetails);

		// ok, it's time to encode the URL for proper transmission
		try {
			encodedUrl = Functions.encodeURL(urlS);
		} catch (MalformedURLException e) {
			String message = "URL: " + urlS + "\r\nMalformedURLException: " + e.getMessage();
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
			return generateErrorEntryList(message, fields, debug, context, tableName, encodedUrl, transid);
		} catch (URISyntaxException e) {
			String message = "URL: " + urlS + "\r\nURISyntaxException: " + e.getMessage();
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
			return generateErrorEntryList(message, fields, debug, context, tableName, encodedUrl, transid);
		}

		try {
			url = new URL(encodedUrl);
		} catch (MalformedURLException e) {
			String message = "URL: " + encodedUrl + "\r\nMalformedURLException: " + e.getMessage();
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
			return generateErrorEntryList(message, fields, debug, context, tableName, encodedUrl, transid);
		}

		if (!encodedUrl.startsWith("http")) {
			String message = encodedUrl;
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
			return generateErrorEntryList(message, fields, debug, context, tableName, encodedUrl, transid);
		}

		if (remedyServer) {
			urlRoot = url.getProtocol() + "://" + url.getHost();

			if (url.getPort() > 0)
				urlRoot += ":" + url.getPort();

			urlRoot += "/";
		}

		if (debug) {
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "URL: " + urlS);
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO,
					"URL-Encoded: " + encodedUrl);
		}

		return null;
	}

	private List<Entry> makeRestCall(ARPluginContext context, String tableName, List<ARVendorField> fields,
			String method, String body, int successResponseCode, Map<String, Object> fieldNameValueMap, long transid)
			throws ARException {
		webConn = null;

		List<Entry> setupEntry = setupWebConnection(context, method, fields, tableName, fieldNameValueMap, transid);

		if (setupEntry != null)
			return setupEntry;

		int bodySize = 0;

		if (body != null && !body.isEmpty()) {
			body = body.trim();

			if (body.startsWith("{") || body.startsWith("[")) { // JSON
				if (debug)
					Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO,
							"Validating body is valid JSON");
				try {

					if (body.startsWith("{")) // JSON Object
						new JSONObject(body);
					if (body.startsWith("[")) // JSON Array
						new JSONArray(body);

					if (debug)
						Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO,
								"Body JSON is valid");
				} catch (JSONException e) {
					String message = "Exception parsing JSON Body: " + e.getMessage();
					Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
					return generateErrorEntryList(message, fields, debug, context, tableName, encodedUrl, transid);
				}
			} else if (body.startsWith("<")) { // XML
				try {
					// Turn xml string into a document
					Document document = DocumentBuilderFactory.newInstance().newDocumentBuilder()
							.parse(new InputSource(new ByteArrayInputStream(body.getBytes("utf-8"))));
					// Remove whitespaces outside tags
					XPath xPath = XPathFactory.newInstance().newXPath();
					NodeList nodeList = (NodeList) xPath.evaluate("//text()[normalize-space()='']", document,
							XPathConstants.NODESET);

					for (int i = 0; i < nodeList.getLength(); ++i) {
						Node node = nodeList.item(i);
						node.getParentNode().removeChild(node);
					}
				} catch (Exception e) {
					String message = "Exception parsing XML Body: " + e.getMessage();
					Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
					return generateErrorEntryList(message, fields, debug, context, tableName, encodedUrl, transid);
				}
			}

			try {
				if (!body.contains("[\r]")) {
					//Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "Replace");
					body = body.replaceAll("[\\n]", "\r\n");
				}

				if (debug)
					Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO,
							"Getting connection OutputStream");
				if (debug) {
					String tmpBody = new String(body);
					if (tmpBody.contains("byteListLoc"))
						tmpBody = body.replaceAll("byteListLoc.*?byteListLoc", "...attachment contents...");
					Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "Adding Body: " + tmpBody);
				}
				List<byte[]> outputList = new ArrayList<>();
				if (body.contains("byteListLoc")) {
					String[] tokens = body.split("byteListLoc");
					for (String s : tokens) {
						//System.out.println("TOKEN: " + s + ": ENDTOKEN");
						int i = -1;
						try {
							i = Integer.parseInt(s);
						} catch (Exception e) {}
						if (i > -1) {
							bodySize += attByteList.get(i).length;
							outputList.add(attByteList.get(i));
						} else {
							byte[] array = s.getBytes(StandardCharsets.UTF_8);
							bodySize += array.length;
							outputList.add(array);
						}
					}
				} else {
					outputList.add(body.getBytes(StandardCharsets.UTF_8));
				}
				if (debug)
					Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO,
							"Adding Header: Content-Length: " + bodySize);
				webConn.setRequestProperty("Content-Length", Integer.toString(bodySize));
				
				OutputStream os = webConn.getOutputStream();
				for (byte[] array : outputList) {
					os.write(array);
				}

				os.flush();
				os.close();
			} catch (IOException e) {
				String message = "Exception adding body to connection: " + e.getMessage();
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
				return generateErrorEntryList(message, fields, debug, context, tableName, encodedUrl, transid);
			}
		} else {
			if (debug)
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO,
						"Adding Header: Content-Length: " + bodySize);
			webConn.setRequestProperty("Content-Length", Integer.toString(bodySize));
		}

		try {
			if (debug)
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "Executing request");
			webConn.connect();
		} catch (IOException e) {
			String message = "IOException connecting to URL: " + e.getMessage();
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
			return generateErrorEntryList(message, fields, debug, context, tableName, encodedUrl, transid);
		}

		String responseMessage = null;
		try {
			responseCode = ((HttpURLConnection) webConn).getResponseCode();
			responseMessage = ((HttpURLConnection) webConn).getResponseMessage();
		} catch (IOException e) {
			String message = "IOException getting response: " + e.getMessage();
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
			return generateErrorEntryList(message, fields, debug, context, tableName, encodedUrl, transid);
		}

		if (responseCode != successResponseCode && successResponseCode != 0 && !allowAllReturnCodes) {
			if (debug) {
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR,
						"Failed: HTTP error code : " + responseCode);
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR,
						"Method: " + method);
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, "URL: " + urlS);
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR,
						"Error Message: " + responseMessage);
			}
			String message = "Failed : HTTP error code : " + responseCode + " Method: " + method + " - Response Message: "
					+ responseMessage + " - URL: " + urlS;
			return generateErrorEntryList(message, fields, debug, context, tableName, encodedUrl, transid);
		}

		if (debug)
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO,
					"Request Complete: Return Code: " + responseCode);

		return null;
	}

	private AttachmentValue getAttachment(ARPluginContext context, String fileName, String urlS,
			List<ARVendorField> fields, String tableName, long transid) throws ARException {
		if (debug)
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO,
					"Initializing getAttachment call");

		List<Entry> tableMetaDataList = getTableMetaData(context, tableName, fields, null, null, false, transid);

		if (tableMetaDataList != null) {
			String message = Functions.getErrorString(fields, tableMetaDataList);
			throw new ARException(Constants.AR_RETURN_ERROR, ARErrors.AR_PLUGIN_ERROR_UNKOWN, message);
		}

		webConn = null;

		if (urlS.startsWith("/")) {
			try {
				URL tempURL = new URL(tableURL);
				String urlRoot = tempURL.getProtocol() + "://" + tempURL.getHost();

				if (tempURL.getPort() > 0)
					urlRoot += ":" + tempURL.getPort();

				urlRoot += "/";
				urlS = urlRoot + urlS;
			} catch (MalformedURLException e) {}
		}
		
		encodedUrl = urlS;

		try {
			url = new URL(encodedUrl);
		} catch (MalformedURLException e) {
			String message = "URL: " + encodedUrl + System.lineSeparator() + "MalformedURLException during getAttachment: " + e.getMessage();
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
			return null;
		}

		String method = "GET";

		List<Entry> setupEntry = setupWebConnection(context, method, fields, tableName, null, transid);

		if (setupEntry != null)
			return null;

		try {
			if (debug)
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "Executing request");
			webConn.connect();
		} catch (IOException e) {
			String message = "IOException connecting to URL during getAttachment: " + e.getMessage();
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
			return null;
		}

		int responseCode = 0;
		String responseMessage = null;
		try {
			responseCode = ((HttpURLConnection) webConn).getResponseCode();
			responseMessage = ((HttpURLConnection) webConn).getResponseMessage();
		} catch (IOException e) {
			String message = "IOException getting response during getAttachment: " + e.getMessage();
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
			return null;
		}

		if (responseCode != HttpURLConnection.HTTP_OK) {
			if (debug) {
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR,
						"Failed : HTTP error code : " + responseCode);
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR,
						"Method: " + method);
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, "URL: " + urlS);
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR,
						"Error Message: " + responseMessage);
			}
			return null;
		}

		if (debug)
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "Request Complete");

		InputStream is = null;

		byte[] bytes = null;
		try {
			try {
				is = webConn.getInputStream();
			} catch (IOException e) {
				is = ((HttpURLConnection) webConn).getErrorStream();
			}
			ByteArrayOutputStream buffer = new ByteArrayOutputStream();

			int nRead;
			byte[] data = new byte[4096];

			while ((nRead = is.read(data, 0, data.length)) != -1) {
				buffer.write(data, 0, nRead);
			}

			buffer.flush();

			bytes = buffer.toByteArray();
		} catch (IOException e) {
			String message = "IOException getting web response during getAttachment: " + e.getMessage();
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
			return null;
		} finally {
			if (is != null) {
				try {
					is.close();
				} catch (IOException e) {
					Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR,
							"Error closing InputStream during getAttachment: " + e.getMessage());
					return null;
				}
			}
		}

		if (bytes != null) {
			AttachmentValue av = new AttachmentValue(bytes);
			av.setName(fileName);
			av.setOriginalSize(bytes.length);

			return av;
		} else {
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, "Returning NULL bytes");
			return null;
		}
	}

	private Entry manageCacheEntry(ARPluginContext context, String url, String body, String tableName, String entryID,
			String entryType, String user, Entry entry, String action, boolean raw) {
		checkConnection(context);

		if (entry != null)
			entryID = entry.getEntryId();

		if (action.equals("Add")) {
			if (body != null) {
				body = body.replaceAll("[^a-zA-Z0-9]", "");
				url += body;
				if (url.length() >= 4000)
					url = url.substring(0, 4000);
			}
			Entry createEntry = new Entry();
			if (url != null)
				createEntry.put(APLConstants.fidCacheURL, new Value(url));
			if (tableName != null)
				createEntry.put(APLConstants.fidCacheTableName, new Value(tableName));
			if (entryID != null)
				createEntry.put(APLConstants.fidCacheEntryID, new Value(entryID));
			if (entryType != null)
				createEntry.put(APLConstants.fidCacheEntryType, new Value(entryType));
			if (user != null)
				createEntry.put(APLConstants.fidCacheUser, new Value(user));

			try {
				createEntry.put(APLConstants.fidCacheEntryObject, new Value(Functions.serializeEntryObject(entry)));
			} catch (Exception e) {
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR,
						"Error converting entry to string: " + e.getMessage());
				return null;
			}

			try {
				conn.createEntry(cacheForm, createEntry);
			} catch (ARException e) {
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR,
						"Error creating CacheEntry: " + e.getMessage());
				return null;
			}
		} else if (action.equals("Get")) {
			QualifierInfo qual = Functions.getQualifierInfo(APLConstants.fidCacheTableName, tableName,
					RelationalOperationInfo.AR_REL_OP_EQUAL);

			if (entryID != null)
				qual = new QualifierInfo(QualifierInfo.AR_COND_OP_AND, qual, Functions.getQualifierInfo(
						APLConstants.fidCacheEntryID, entryID, RelationalOperationInfo.AR_REL_OP_EQUAL));
			if (url != null)
				qual = new QualifierInfo(QualifierInfo.AR_COND_OP_AND, qual, Functions
						.getQualifierInfo(APLConstants.fidCacheURL, url, RelationalOperationInfo.AR_REL_OP_EQUAL));
			if (entryType != null)
				qual = new QualifierInfo(QualifierInfo.AR_COND_OP_AND, qual, Functions.getQualifierInfo(
						APLConstants.fidCacheEntryType, entryType, RelationalOperationInfo.AR_REL_OP_EQUAL));
			if (user != null)
				qual = new QualifierInfo(QualifierInfo.AR_COND_OP_AND, qual, Functions
						.getQualifierInfo(APLConstants.fidCacheUser, user, RelationalOperationInfo.AR_REL_OP_EQUAL));

			int[] fieldArray = { APLConstants.fidCacheEntryObject };

			List<Entry> entryList = null;
			try {
				entryList = conn.getListEntryObjects(cacheForm, qual, 0, 0, null, fieldArray, false, null);
			} catch (ARException e) {
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR,
						"Error getting CacheEntry: " + e.getMessage());
				return null;
			}

			if (entryList != null) {
				for (Entry returnedEntry : entryList) {
					if (raw) {
						return returnedEntry;
					} else {
						try {
							return Functions.deserializeEntryObject(context,
									returnedEntry.get(APLConstants.fidCacheEntryObject).toString());
						} catch (Exception e) {
							Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR,
									"Error converting String to Entry Object CacheEntry: " + e.getMessage());
							return null;
						}
					}
				}
			}
		} else if (action.equals("Delete")) {
			if (entryID != null && !entryID.isEmpty()) {
				try {
					conn.deleteEntry(cacheForm, entryID, 0);
				} catch (ARException e) {
					Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR,
							"Error deleting CacheEntry: " + e.getMessage());
					return null;
				}
			} else {
				try {
					conn.executeProcess(
							"Application-Query-Delete-Entry \"" + cacheForm + "\" 'TableName' = \"" + tableName + "\"");
				} catch (ARException e) {
					Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR,
							"Error deleting CacheEntry: " + e.getMessage());
					return null;
				}
			}
		}

		return null;
	}

	private List<Entry> getListCacheEntry(ARPluginContext context, String tableName, String url, String body,
			String entryType, boolean light) {
		checkConnection(context);

		QualifierInfo qual = null;
		if (tableName == null) {
			qual = Functions.getQualifierInfo(APLConstants.fidCacheEntryType, entryType,
					RelationalOperationInfo.AR_REL_OP_EQUAL);
		} else {
			if (body != null) {
				body = body.replaceAll("[^a-zA-Z0-9]", "");
				url += body;
			}
			QualifierInfo tableQual = Functions.getQualifierInfo(APLConstants.fidCacheTableName, tableName,
					RelationalOperationInfo.AR_REL_OP_EQUAL);
			QualifierInfo urlQual = Functions.getQualifierInfo(APLConstants.fidCacheURL, url,
					RelationalOperationInfo.AR_REL_OP_EQUAL);
			QualifierInfo entryTypeQual = Functions.getQualifierInfo(APLConstants.fidCacheEntryType, entryType,
					RelationalOperationInfo.AR_REL_OP_EQUAL);
			QualifierInfo qual1 = new QualifierInfo(QualifierInfo.AR_COND_OP_AND, tableQual, urlQual);
			qual = new QualifierInfo(QualifierInfo.AR_COND_OP_AND, qual1, entryTypeQual);
		}

		List<Entry> entryList = null;
		try {
			if (light)
				entryList = conn.getListEntryObjects(cacheForm, qual, 0, 0, null, APLConstants.cacheFieldsLight, false,
						null);
			else
				entryList = conn.getListEntryObjects(cacheForm, qual, 0, 0, null, APLConstants.cacheFields, false,
						null);
		} catch (ARException e) {
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR,
					"Error getting CacheEntry: " + e.getMessage());
			return null;
		}

		return entryList;
	}

	private void checkConnection(ARPluginContext context) {
		if (conn == null)
			initializeARConnection(context);
		else {
			try {
				conn.login();
			} catch (ARException e) {
				initializeARConnection(context);
			}
		}
	}

	private void initializeARConnection(ARPluginContext context) {
		if (context != null) {
			String rasHash = context.getARConfigEntry("Remedy-App-Service-Password");

			if (rasHash != null && !rasHash.isEmpty()) {
				String user = "Remedy Application Service";
				String pass = new ARUtilEgcp().GCEUtilApp(rasHash);

				if (pass != null && !pass.isEmpty())
					conn = new ARServerUser(user, pass, null, remedyServerName);
				else {
					Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR,
							"Error initializing connection to Remedy server, unable to get RAS Password from config");
					conn = null;
					return;
				}
			} else {
				conn = new ARServerUser("APL", "", null, "VWinServer");
				conn.setPort(1976);
			}
		}
		if (remedyServerPort > 0)
			conn.setPort(remedyServerPort);

		try {
			conn.login();
		} catch (ARException e) {
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR,
					"Error initializing connection to Remedy server: " + e.getMessage());
		}
	}

	private List<Entry> setupWebConnection(ARPluginContext context, String method, List<ARVendorField> fields,
			String tableName, Map<String, Object> fieldNameValueMap, long transid) throws ARException {
		if (encodedUrl.startsWith("https") && bypassCertCheck) {
			if (debug)
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO,
						"Bypassing SSL Certification Check");
			try {
				SSLContext sc = SSLContext.getInstance("TLS");
				sc.init(null, trustAllCerts, new java.security.SecureRandom());
				HttpsURLConnection.setDefaultSSLSocketFactory(sc.getSocketFactory());
				HttpsURLConnection.setDefaultHostnameVerifier(allHostsValid);
			} catch (NoSuchAlgorithmException e) {
				String message = "NoSuchAlgorithmException while installing TrustManager: " + e.getMessage();
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
				return null;
			} catch (KeyManagementException e) {
				String message = "KeyManagementException while installing TrustManager: " + e.getMessage();
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
				return null;
			}
		}

		if (!proxyUser.isEmpty() && !bypassProxy) {
			Authenticator authenticator = new Authenticator() {

				public PasswordAuthentication getPasswordAuthentication() {
					return (new PasswordAuthentication(proxyUser, proxyPass.toCharArray()));
				}
			};
			Authenticator.setDefault(authenticator);
		} else if (!tableBasicUser.isEmpty()) {
			basicAuth = new BasicAuthenticator(tableBasicUser, tableBasicPass);
			Authenticator.setDefault(basicAuth);
		}

		try {
			if (tableTimeoutS != null && !tableTimeoutS.isEmpty()) {
				int tableTimeout = Integer.parseInt(tableTimeoutS.trim());
				timeout = tableTimeout;
			}
		} catch (Exception e) {
		}

		if (debug) {
			String debugOut = "Starting processing of request with timeout of " + timeout + " seconds";
			if (!proxyServer.isEmpty() && !bypassProxy)
				debugOut += " through Proxy Server " + proxyServer + ":" + proxyPort;
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, debugOut);
		}

		try {
			if (proxyServer.isEmpty() || bypassProxy) {
				webConn = url.openConnection();
			} else {
				webConn = url.openConnection(proxy);
			}
		} catch (SocketTimeoutException e) {
			String message = "SocketTimeoutException opening URL: " + e.getMessage();
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
			return null;
		} catch (SecurityException e) {
			String message = "SecurityException opening URL: " + e.getMessage();
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
			return null;
		} catch (IOException e) {
			String message = "IOException opening URL: " + e.getMessage();
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
			return null;
		}

		webConn.setConnectTimeout(timeout * 1000);
		webConn.setReadTimeout(timeout * 1000);

		// time to set the keystore

		if (webConn instanceof HttpsURLConnection && keyStoreFile != null && !keyStoreFile.isEmpty()) {
			KeyManagerFactory keyManagerFactory = null;
			try {
				keyManagerFactory = KeyManagerFactory.getInstance(keyStoreAlgorithm);
			} catch (NoSuchAlgorithmException e) {
				String message = "NoSuchAlgorithmException establishing KeyManegerFactory: " + e.getMessage();
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
				return null;
			}
			KeyStore keyStore = null;
			try {
				keyStore = KeyStore.getInstance(keyStoreFormat);
			} catch (KeyStoreException e) {
				String message = "KeyStoreException establishing KeyStore: " + e.getMessage();
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
				return null;
			}

			InputStream keyInput = null;
			try {
				keyInput = new FileInputStream(keyStoreFile);
			} catch (FileNotFoundException e) {
				String message = "FileNotFoundException opening keyFile: " + e.getMessage();
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
				return null;
			}
			try {
				keyStore.load(keyInput, keyStorePass.toCharArray());
			} catch (NoSuchAlgorithmException e) {
				String message = "NoSuchAlgorithmException loading keyStore: " + e.getMessage();
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
				return null;
			} catch (CertificateException e) {
				String message = "CertificateException loading keyStore: " + e.getMessage();
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
				return null;
			} catch (IOException e) {
				String message = "IOException loading keyStore: " + e.getMessage();
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
				return null;
			}
			try {
				keyInput.close();
			} catch (IOException e) {
				String message = "IOException closing keyFile: " + e.getMessage();
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
				return null;
			}

			try {
				keyManagerFactory.init(keyStore, keyStorePass.toCharArray());
			} catch (UnrecoverableKeyException e) {
				String message = "UnrecoverableKeyException initializing keyManagerFactory: " + e.getMessage();
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
				return null;
			} catch (KeyStoreException e) {
				String message = "KeyStoreException initializing keyManagerFactory: " + e.getMessage();
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
				return null;
			} catch (NoSuchAlgorithmException e) {
				String message = "NoSuchAlgorithmException initializing keyManagerFactory: " + e.getMessage();
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
				return null;
			}

			SSLContext sslContext = null;
			try {
				sslContext = SSLContext.getInstance(keyStoreProtocol);
			} catch (NoSuchAlgorithmException e) {
				String message = "NoSuchAlgorithmException setting sslContext to " + keyStoreProtocol + ": "
						+ e.getMessage();
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
				return null;
			}
			try {
				sslContext.init(keyManagerFactory.getKeyManagers(), null, new SecureRandom());
			} catch (KeyManagementException e) {
				String message = "KeyManagementException ititializing sslContext: " + e.getMessage();
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
				return null;
			}

			((HttpsURLConnection) webConn).setSSLSocketFactory(sslContext.getSocketFactory());
		}

		Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "Using Method: " + method);

		if (remedyServer) {
			if (!tableBasicUser.isEmpty()) {
				String urlSL = urlRoot + "api/jwt/login";

				if (debug)
					Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO,
							"Getting Remedy Authentication Token from URL '" + urlSL + "'");

				try {
					URL urlL = new URL(urlSL);
					if (proxyServer.isEmpty() || bypassProxy) {
						webConn = urlL.openConnection();
					} else {
						webConn = urlL.openConnection(proxy);
					}
					((HttpURLConnection) webConn).setRequestMethod("POST");
					webConn.setRequestProperty("Content-Type", "application/x-www-form-urlencoded");

					webConn.setUseCaches(false);
					webConn.setDoInput(true);
					webConn.setDoOutput(true);

					String urlParameters = "username=" + tableBasicUser + "&password=" + tableBasicPass;

					DataOutputStream wr = new DataOutputStream(webConn.getOutputStream());
					wr.writeBytes(urlParameters);
					wr.flush();
					wr.close();

					int webResponseCode = 0;

					webResponseCode = ((HttpURLConnection) webConn).getResponseCode();

					if (webResponseCode != HttpURLConnection.HTTP_OK) {
						Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR,
								"Error getting Remedy connection Token");
						Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR,
								"HTTP error code: " + webResponseCode);
						Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR,
								"URL: " + urlSL);
						String message = "Error getting Remedy connection Token - HTTP error code: " + webResponseCode
								+ " URL: " + urlSL;
						return generateErrorEntryList(message, fields, debug, context, tableName, encodedUrl, transid);
					}

					InputStream is = null;
					
					try {
						is = webConn.getInputStream();
					} catch (IOException e) {
						is = ((HttpURLConnection) webConn).getErrorStream();
					}
					
					BufferedReader rd = new BufferedReader(new InputStreamReader(is));
					remedyToken = rd.readLine();
					rd.close();

					remedyToken = "AR-JWT " + remedyToken;

					if (debug)
						Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO,
								"Token retrieved");
				} catch (MalformedURLException e) {
					String message = "MalformedURLException getting Token: " + e.getMessage();
					Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
					return generateErrorEntryList(message, fields, debug, context, tableName, encodedUrl, transid);
				} catch (ProtocolException e) {
					String message = "ProtocolException: getting Token: " + e.getMessage();
					Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
					return generateErrorEntryList(message, fields, debug, context, tableName, encodedUrl, transid);
				} catch (IOException e) {
					String message = "IOException: getting Token: " + e.getMessage();
					Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
					return generateErrorEntryList(message, fields, debug, context, tableName, encodedUrl, transid);
				} finally {
					if (webConn != null) {
						((HttpURLConnection) webConn).disconnect();
					}
				}
			}

			if (debug)
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO,
						"Adding Header: Authorization: " + remedyToken);
			try {
				if (proxyServer.isEmpty() || bypassProxy) {
					webConn = url.openConnection();
				} else {
					webConn = url.openConnection(proxy);
				}
			} catch (IOException e) {
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO,
						"IOException opening the URL: " + e.getMessage());
			}
			webConn.setRequestProperty("Authorization", remedyToken);
		} else {
			if (!tableBasicUser.isEmpty()) {
				String authString = Functions.base64EncodeString(tableBasicUser + ":" + tableBasicPass);
				if (debug)
					Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO,
							"Adding Header: Authorization: Basic " + tableBasicUser + ":******");
				webConn.setRequestProperty("Authorization", "Basic " + authString);
			}
		}

		try {
			((HttpURLConnection) webConn).setRequestMethod(method);
		} catch (ProtocolException e) {
			allowMethods(method);
			try {
				((HttpURLConnection) webConn).setRequestMethod(method);
			} catch (ProtocolException e2) {
				String message = "ProtocolException setting Method for URL: " + e2.getMessage();
				Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, message);
				return generateErrorEntryList(message, fields, debug, context, tableName, encodedUrl, transid);
			}
		}

		webConn.setDoOutput(true);
		webConn.setDoInput(true);
		HttpURLConnection.setFollowRedirects(true);
		HttpsURLConnection.setFollowRedirects(true);

		if (tableCustomHeader == null || tableCustomHeader.isEmpty()) {
			webConn.setRequestProperty("Accept", "*/*");
		} else {
			String headerTemplate = tableCustomHeader;
			if (headerTemplate != null) {
				if (fieldNameValueMap != null) {
					headerTemplate = Functions.replaceTemplateValues(tableCustomHeader, fieldNameValueMap,
							fieldOptionalList, aplFieldMap, templateFieldDelimiter);
				}

				String[] tokens = headerTemplate.split("\n");
				for (int i = 0; i < tokens.length; i++) {
					int firstColon = tokens[i].indexOf(":");
					if (firstColon > 0) {
						String header = tokens[i].substring(0, firstColon).trim();
						String value = tokens[i].substring(firstColon + 1).trim();
						if (debug)
							Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO,
									"Adding Header: " + header + ": " + value);
						webConn.setRequestProperty(header, value);
					} else {
						if (debug) {
							Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO,
									"Error adding header, not in the format header:value");
							Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO,
									tokens[i]);
						}
					}
				}
			}
		}

		return null;
	}

	private final TrustManager[] trustAllCerts = new TrustManager[] { new X509TrustManager() {
		public java.security.cert.X509Certificate[] getAcceptedIssuers() {
			return null;
		}

		public void checkClientTrusted(X509Certificate[] certs, String authType) {
		}

		public void checkServerTrusted(X509Certificate[] certs, String authType) {
		}
	} };

	private final HostnameVerifier allHostsValid = new HostnameVerifier() {
		public boolean verify(String hostname, SSLSession session) {
			return true;
		}
	};

	private static void allowMethods(String... methods) {
		try {
			java.lang.reflect.Field methodsField = HttpURLConnection.class.getDeclaredField("methods");

			java.lang.reflect.Field modifiersField = java.lang.reflect.Field.class.getDeclaredField("modifiers");
			modifiersField.setAccessible(true);
			modifiersField.setInt(methodsField, methodsField.getModifiers() & ~java.lang.reflect.Modifier.FINAL);

			methodsField.setAccessible(true);

			String[] oldMethods = (String[]) methodsField.get(null);
			Set<String> methodsSet = new LinkedHashSet<>(Arrays.asList(oldMethods));
			methodsSet.addAll(Arrays.asList(methods));
			String[] newMethods = methodsSet.toArray(new String[0]);

			methodsField.set(null/* static field */, newMethods);
		} catch (NoSuchFieldException | IllegalAccessException e) {
			throw new IllegalStateException(e);
		}
	}
}
