/*
	APL Restful ARDBC Plugin
    Copyright 2015-2018  LJ LongWing - A Programming Legacy (http://remedylegacy.com)

	Licensed under the Apache License, Version 2.0 (the "License");
	you may not use this file except in compliance with the License.
	You may obtain a copy of the License at

		http://www.apache.org/licenses/LICENSE-2.0

	Unless required by applicable law or agreed to in writing, software
	distributed under the License is distributed on an "AS IS" BASIS,
	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
	See the License for the specific language governing permissions and
	limitations under the License.
*/

package com.apl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import com.bmc.arsys.pluginsvr.plugins.ARVendorField;

public class XMLParser extends DefaultHandler {
	private String elementName = "";
	private String elementValue = "";
	private String tableRecordIndicator = "";
	private String elementPrefix = "";
	
	private int currentLevel;

	private List<Map<String, String>> recordElementList = new ArrayList<Map<String, String>>();
	private List<Integer> skipLevelList = new ArrayList<Integer>();
	private List<String> fieldNameList = new ArrayList<String>();

	private Map<String, String> globalElementMap = new HashMap<String, String>();
	private Map<String, String> recordElementMap;
	private Map<String, List<String>> arrayElementMap = new TreeMap<String, List<String>>();

	private boolean inRecord = false;

	public XMLParser(String tableRecordIndicator, List<Integer> skipLevelList, List<ARVendorField> fieldList) {
		super();
		if (fieldList != null) {
			for (ARVendorField field : fieldList) {
				this.fieldNameList.add(field.getName());
			}
		}
		if (tableRecordIndicator != null)
			this.tableRecordIndicator = tableRecordIndicator;
		if (skipLevelList != null)
			this.skipLevelList = skipLevelList;
	}

	public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException {
		currentLevel++;

		if (!skipLevelList.contains(currentLevel)) {
			elementName = qName.trim();

			elementPrefix = Functions.addElementPrefix(elementPrefix, elementName);
			if (elementPrefix.substring(0, elementPrefix.length() - 1).equals(tableRecordIndicator)) {
				inRecord = true;
				recordElementMap = new HashMap<String, String>();
			}
			for (int i = 0; i < attributes.getLength(); i++) {
				String attName = attributes.getLocalName(i).trim();
				String attValue = attributes.getValue(i).trim();
				if (inRecord) {
					recordElementMap = populateMap(recordElementMap, elementPrefix + attName, attValue);
				} else {
					globalElementMap = populateMap(globalElementMap, elementPrefix + attName, attValue);
				}
			}
			elementValue = "";
		}
	}

	public void endElement(String uri, String localName, String qName) throws SAXException {
		if (!skipLevelList.contains(currentLevel)) {
			elementName = qName.trim();

			String noAtt = elementPrefix.substring(0, elementPrefix.length() - 1);

			if (!elementValue.isEmpty()) {
				if (inRecord) {
					recordElementMap = populateMap(recordElementMap, noAtt, elementValue);
				} else {
					globalElementMap = populateMap(globalElementMap, noAtt, elementValue);
				}
			}
			elementValue = "";

			if (elementPrefix.substring(0, elementPrefix.length() - 1).equals(tableRecordIndicator)) {
				inRecord = false;
				recordElementList.add(recordElementMap);
			}
			elementPrefix = Functions.removeElementSuffix(elementPrefix, elementName);
		}
		currentLevel--;
	}

	public void characters(char ch[], int start, int length) throws SAXException {
		String elementCharacters = new String(ch, start, length);
		elementCharacters = elementCharacters.trim();
		if (!(globalElementMap.containsKey(elementName) && elementCharacters.isEmpty())) {
			elementValue += elementCharacters;
		}
	}

	private Map<String, String> populateMap(Map<String, String> map, String element, String value) {
		value = value.trim();
		if (!map.containsKey(element)) {
			map.put(element, value);
		} else {
			List<String> currentValueList = arrayElementMap.get(element);
			if (currentValueList == null) {
				String curValue = map.get(element);
				currentValueList = new ArrayList<String>();
				currentValueList.add(curValue);
			}
			currentValueList.add(value);
			arrayElementMap.put(element, currentValueList);
		}

		return map;
	}

	public Map<String, String> getGlobalElementMap(String arrayDelimiter) {
		List<Map<String, String>> mapList = Functions.processArrays(arrayElementMap, globalElementMap, recordElementMap, fieldNameList , arrayDelimiter);
		if (mapList != null && mapList.get(0) != null)
			globalElementMap = mapList.get(0);
		return globalElementMap;
	}

	public Map<String, String> getRecordElementMap(String arrayDelimiter) {
		List<Map<String, String>> mapList = Functions.processArrays(arrayElementMap, globalElementMap, recordElementMap, fieldNameList, arrayDelimiter);
		if (mapList != null && mapList.get(1) != null)
			recordElementMap = mapList.get(1);
		return recordElementMap;
	}

	public List<Map<String, String>> getRecordElementList() {
		return recordElementList;
	}
}
